"""
Implements the Decorator HOP task using the TDC oracles [1].

This task is inherited from the GuacaMol benchmark [2], and
consists of maximizing the similarity to a certain SMILES,
"while excluding specific SMARTS patterns" [1]. We recommend
you cite both references when using this task.

References
----------
[1] Artificial intelligence foundation for therapeutic science.
    Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022). https://doi.org/10.1038/s41589-022-01131-2
[2] GuacaMol: benchmarking models for de novo molecular design.
    Brown, N. et al.  J Chem Inf Model 59 (2019).
    https://pubs.acs.org/doi/10.1021/acs.jcim.8b00839
"""

from __future__ import annotations

from typing import Literal

import numpy as np
import selfies as sf

from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.chemistry.tdc_black_box import TDCBlackBox
from poli.core.chemistry.tdc_problem import TDCProblem
from poli.core.problem import Problem
from poli.core.util.chemistry.string_to_molecule import translate_smiles_to_selfies
from poli.core.util.seeding import seed_numpy, seed_python


class DecoHopBlackBox(TDCBlackBox):
    """
    The "Deco HOP" oracle from TDC [1].

    This task is inherited from the GuacaMol benchmark [2], and
    consists of maximizing the similarity to a certain SMILES,
    "while excluding specific SMARTS patterns" [1]. We recommend
    you cite both references when using this task.

    Parameters
    ----------
    string_representation : Literal["SMILES", "SELFIES"], optional
        A string (either "SMILES" or "SELFIES") specifying which
        molecule representation you plan to use.
    batch_size : int, optional
        The batch size for simultaneous execution, by default None.
    parallelize : bool, optional
        Flag indicating whether to parallelize execution, by default False.
    num_workers : int, optional
        The number of workers for parallel execution, by default None.
    evaluation_budget:  int, optional
        The maximum number of function evaluations. Default is infinity.
    force_isolation: bool, optional
        Whether to force the isolation of the black box. Default is False.

    Attributes
    ----------
    oracle_name : str
        The name of the oracle.

    Methods
    -------
    __init__(self, string_representation, force_isolation, batch_size=None, parallelize=False, num_workers=None, evaluation_budget=float("inf"))
        Creates a DecoHopBlackBox instance.

    References
    ----------
    [1] Artificial intelligence foundation for therapeutic science.
        Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022).
        https://doi.org/10.1038/s41589-022-01131-2
    [2] GuacaMol: benchmarking models for de novo molecular design.
        Brown, N. et al.  J Chem Inf Model 59 (2019).
        https://pubs.acs.org/doi/10.1021/acs.jcim.8b00839
    """

    def __init__(
        self,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        alphabet: list[str] | None = None,
        max_sequence_length: int = np.inf,
        force_isolation: bool = False,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
    ):
        super().__init__(
            oracle_name="Deco Hop",
            string_representation=string_representation,
            alphabet=alphabet,
            max_sequence_length=max_sequence_length,
            force_isolation=force_isolation,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="deco_hop",
            max_sequence_length=self.max_sequence_length,
            aligned=False,
            fixed_length=False,
            deterministic=True,  # ?
            alphabet=self.alphabet,  # TODO: add alphabet once we settle for one for SMLIES/SELFIES.
            log_transform_recommended=False,
            discrete=True,
            padding_token="",
        )


class DecoHopProblemFactory(AbstractProblemFactory):
    """
    Factory class for creating Decorator Hop problems.

    We recommend you cite [1, 2] when using this task.

    Methods
    ------
    create:
        Creates an Decorator Hop problem, containing a black box
        and an initial value x0 (taken from the documentation of TDC).

    References
    ----------
    [1] Artificial intelligence foundation for therapeutic science.
        Huang, K., Fu, T., Gao, W. et al.  Nat Chem Biol 18, 1033-1036 (2022).
        https://doi.org/10.1038/s41589-022-01131-2
    [2] GuacaMol: benchmarking models for de novo molecular design.
        Brown, N. et al.  J Chem Inf Model 59 (2019).
        https://pubs.acs.org/doi/10.1021/acs.jcim.8b00839
    """

    def create(
        self,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
        alphabet: list[str] | None = None,
        max_sequence_length: int = np.inf,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Creates a Decorator Hop problem.

        Parameters
        ----------
        string_representation : str, optional
            The string representation of the molecules. Must be either 'SMILES' or 'SELFIES'. Default is 'SMILES'.
        seed : int, optional
            Seed for random number generators. If None, no seed is set.
        batch_size : int, optional
            Number of molecules to process in parallel. If None, the default batch size is used.
        parallelize : bool, optional
            Whether to parallelize the docking process. Default is False.
        num_workers : int, optional
            Number of worker processes to use for parallelization. If None, the number of available CPU cores is used.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        force_isolation: bool, optional
            Whether to force the isolation of the black box. Default is False.

        Returns
        -------
        problem : Problem
            A problem instance containing the black box, and an initial value x0.

        Raises
        ------
        ValueError
            If the string_representation is not 'SMILES' or 'SELFIES'.
        """
        # We start by seeding the RNGs
        if seed is not None:
            seed_numpy(seed)
            seed_python(seed)

        # We check whether the string representation is valid
        if string_representation.upper() not in ["SMILES", "SELFIES"]:
            raise ValueError(
                "Missing required keyword argument: string_representation: str. "
                "String representation must be either 'SMILES' or 'SELFIES'."
            )

        f = DecoHopBlackBox(
            string_representation=string_representation,
            alphabet=alphabet,
            max_sequence_length=max_sequence_length,
            force_isolation=force_isolation,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

        # Initial example (from the TDC docs)
        x0_smiles = "CC(C)(C)[C@H]1CCc2c(sc(NC(=O)COc3ccc(Cl)cc3)c2C(N)=O)C1"
        x0_selfies = translate_smiles_to_selfies([x0_smiles])[0]

        if string_representation.upper() == "SMILES":
            x0 = np.array([list(x0_smiles)])
        else:
            x0 = np.array([list(sf.split_selfies(x0_selfies))])

        deco_hop_problem = TDCProblem(
            black_box=f,
            x0=x0,
        )

        return deco_hop_problem



"""
This module implements a lambda black box:

Once the user has specified a callable f which
takes as input np.ndarray and returns np.ndarray,
this black box uses it for the evaluation.

This is useful when the user has already implemented
a lot of functionality around a callable, and wants to
leverage all that comes with a poli black box: logging,
budget management, etc.
"""

from __future__ import annotations

from typing import Callable

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.black_box_information import BlackBoxInformation


class LambdaBlackBox(AbstractBlackBox):
    """
    A Lambda black box is a black box that uses a callable
    for the evaluation.

    Parameters:
    -----------
    function : Callable[[np.ndarray], np.ndarray]
        The callable that will be used for the evaluation.
    info : BlackBoxInformation, optional
        The information about the black box. If None,
        a default information is used.
    batch_size : int, optional
        The batch size for evaluating the black box function. Default is None.
    parallelize : bool, optional
        Flag indicating whether to evaluate the black box function in parallel.
        Default is False.
    num_workers : int, optional
        The number of workers to use for parallel evaluation. Default is None,
        which uses half of the available CPU cores.
    evaluation_budget : int, optional
        The maximum number of evaluations allowed for the black box function.
        Default is float("inf").
    force_isolation : bool, optional
        Flag indicating whether to force isolation of the black
        box function. In this black box, this flag is ignored.
    """

    def __init__(
        self,
        function: Callable[[np.ndarray], np.ndarray],
        info: BlackBoxInformation | None = None,
        batch_size=None,
        parallelize=False,
        num_workers=None,
        evaluation_budget=float("inf"),
        force_isolation=False,
    ):
        super().__init__(
            batch_size, parallelize, num_workers, evaluation_budget, force_isolation
        )
        self.function = function
        self._information = (
            info
            if info is not None
            else BlackBoxInformation(
                name="lambda_function",
                max_sequence_length=np.inf,
                aligned=False,
                fixed_length=False,
                deterministic=False,
                alphabet=None,
                discrete=True,
            )
        )

    def _black_box(self, x, context=None):
        return self.function(x)

    def get_black_box_info(self):
        return self._information



"""Registers the Super Mario Bros objective function and factory.

The metric being recorded is the number of
jumps Mario makes in a given level. The goal
is, of course, maximizing it while keeping the
level playable.

"""

from pathlib import Path

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.isolation.instancing import get_inner_function
from poli.core.util.seeding import seed_python_numpy_and_torch
from poli.objective_repository.super_mario_bros.information import SMB_ALPHABET

THIS_DIR = Path(__file__).parent.resolve()

# TODO: download the model from the internet
# if it doesn't exist, as well as the simulator
# FIXME: do this after we remove .pt and .jar
# files from the python installation.


class SuperMarioBrosBlackBox(AbstractBlackBox):
    """
    Black box which returns how many jumps were performed
    by Mario in the given level.

    Parameters
    ----------
    max_time : int, optional
        The maximum time for the simulation in seconds, by default 30.
    visualize : bool, optional
        Flag indicating whether to visualize the simulation, by default False.
    batch_size : int, optional
        The batch size for simultaneous execution, by default None.
    parallelize : bool, optional
        Flag indicating whether to parallelize execution, by default False.
    num_workers : int, optional
        The number of workers for parallel execution, by default None.
    evaluation_budget:  int, optional
        The maximum number of function evaluations. Default is infinity.

    Methods
    -------
    _black_box(self, x, context=None)
        Runs the given input x as a latent code
        through the model and returns the number
        of jumps Mario makes in the level. If the
        level is not solvable, returns np.nan.
    """

    def __init__(
        self,
        max_time: int = 30,
        visualize: bool = False,
        value_on_unplayable: float = np.nan,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        """
        Initializes a new instance of the SMBBlackBox class.

        Parameters
        ----------
        batch_size : int, optional
            The batch size for simultaneous execution, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize execution, by default False.
        num_workers : int, optional
            The number of workers for parallel execution, by default None.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        """
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        self.force_isolation = force_isolation
        self.max_time = int(max_time)
        self.visualize = visualize
        self.value_on_unplayable = value_on_unplayable
        _ = get_inner_function(
            isolated_function_name="super_mario_bros__isolated",
            class_name="SMBIsolatedLogic",
            module_to_import="poli.objective_repository.super_mario_bros.isolated_function",
            force_isolation=self.force_isolation,
            alphabet=self.get_black_box_info().alphabet,
            max_time=self.max_time,
            visualize=self.visualize,
            value_on_unplayable=self.value_on_unplayable,
        )

    def _black_box(self, x: np.ndarray, context=None) -> np.ndarray:
        """Computes number of jumps in a given latent code x."""
        inner_function = get_inner_function(
            isolated_function_name="super_mario_bros__isolated",
            class_name="SMBIsolatedLogic",
            module_to_import="poli.objective_repository.super_mario_bros.isolated_function",
            force_isolation=self.force_isolation,
            quiet=True,
            alphabet=self.get_black_box_info().alphabet,
            max_time=self.max_time,
            visualize=self.visualize,
            value_on_unplayable=self.value_on_unplayable,
        )
        return inner_function(x, context)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="super_mario_bros",
            max_sequence_length=14 * 14,
            aligned=True,
            fixed_length=True,
            deterministic=False,
            alphabet=SMB_ALPHABET,
            log_transform_recommended=True,
            discrete=True,
            padding_token=None,
        )


class SuperMarioBrosProblemFactory(AbstractProblemFactory):
    """
    Problem factory for the Super Mario Bros objective function.

    Methods
    -------
    create(...) -> Tuple[SMBBlackBox, np.ndarray, np.ndarray]
        Creates a new instance of the SMBBlackBox class.
    """

    def create(
        self,
        max_time: int = 30,
        visualize: bool = False,
        value_on_unplayable: float = np.nan,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """Creates a new instance of the Super Mario Bros problem.

        Parameters
        ----------
        max_time : int, optional
            The maximum time for the simulation in seconds, by default 30.
        visualize : bool, optional
            Flag indicating whether to visualize the simulation, by default False.
        seed : int, optional
            The seed for the random number generator, by default None.
        batch_size : int, optional
            The batch size for simultaneous execution, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize execution, by default False.
        num_workers : int, optional
            The number of workers for parallel execution, by default None.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        force_isolation: bool, optional
            Flag indicating whether to force isolation inside the black box,
            by default False.

        Returns
        -------
        problem : Problem:
            The problem instance containing the black box and an initial point.
        """
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        f = SuperMarioBrosBlackBox(
            max_time=max_time,
            visualize=visualize,
            value_on_unplayable=value_on_unplayable,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )
        x0 = np.array([["-"] * 14] * 14)
        x0[-1, :] = "X"
        x0 = x0.reshape(1, 14 * 14)

        problem = Problem(f, x0)

        return problem



"""All objective factories and black boxes inside poli.
"""

from pathlib import Path

from .albuterol_similarity.register import (
    AlbuterolSimilarityBlackBox,
    AlbuterolSimilarityProblemFactory,
)
from .aloha.register import AlohaBlackBox, AlohaProblemFactory
from .amlodipine_mpo.register import AmlodipineMPOBlackBox, AmlodipineMPOProblemFactory
from .celecoxib_rediscovery.register import (
    CelecoxibRediscoveryBlackBox,
    CelecoxibRediscoveryProblemFactory,
)
from .deco_hop.register import DecoHopBlackBox, DecoHopProblemFactory
from .dockstring.register import DockstringBlackBox, DockstringProblemFactory
from .drd2_docking.register import DRD2BlackBox, DRD2ProblemFactory
from .drd3_docking.register import DRD3BlackBox, DRD3ProblemFactory
from .ehrlich.register import EhrlichBlackBox, EhrlichProblemFactory
from .ehrlich_holo.register import EhrlichHoloBlackBox, EhrlichHoloProblemFactory
from .fexofenadine_mpo.register import (
    FexofenadineMPOBlackBox,
    FexofenadineMPOProblemFactory,
)
from .foldx_rfp_lambo.register import FoldXRFPLamboBlackBox, FoldXRFPLamboProblemFactory
from .foldx_sasa.register import FoldXSASABlackBox, FoldXSASAProblemFactory
from .foldx_stability.register import (
    FoldXStabilityBlackBox,
    FoldXStabilityProblemFactory,
)
from .foldx_stability_and_sasa.register import (
    FoldXStabilityAndSASABlackBox,
    FoldXStabilityAndSASAProblemFactory,
)
from .gfp_cbas.register import GFPCBasBlackBox, GFPCBasProblemFactory
from .gfp_select.register import GFPSelectionBlackBox, GFPSelectionProblemFactory
from .gsk3_beta.register import GSK3BetaBlackBox, GSK3BetaProblemFactory
from .isomer_c7h8n2o2.register import (
    IsomerC7H8N2O2BlackBox,
    IsomerC7H8N2O2ProblemFactory,
)
from .isomer_c9h10n2o2pf2cl.register import (
    IsomerC9H10N2O2PF2ClBlackBox,
    IsomerC9H10N2O2PF2ClProblemFactory,
)
from .jnk3.register import JNK3BlackBox, JNK3ProblemFactory
from .median_1.register import Median1BlackBox, Median1ProblemFactory
from .median_2.register import Median2BlackBox, Median2ProblemFactory
from .mestranol_similarity.register import (
    MestranolSimilarityBlackBox,
    MestranolSimilarityProblemFactory,
)
from .osimetrinib_mpo.register import (
    OsimetrinibMPOBlackBox,
    OsimetrinibMPOProblemFactory,
)
from .penalized_logp_lambo.register import (
    PenalizedLogPLamboBlackBox,
    PenalizedLogPLamboProblemFactory,
)
from .perindopril_mpo.register import (
    PerindoprilMPOBlackBox,
    PerindoprilMPOProblemFactory,
)
from .ranolazine_mpo.register import RanolazineMPOBlackBox, RanolazineMPOProblemFactory
from .rasp.register import RaspBlackBox, RaspProblemFactory
from .rdkit_logp.register import LogPBlackBox, LogPProblemFactory
from .rdkit_qed.register import QEDBlackBox, QEDProblemFactory
from .rfp_foldx_stability.register import (
    RFPFoldXStabilityBlackBox,
    RFPFoldXStabilityProblemFactory,
)
from .rfp_foldx_stability_and_sasa.register import (
    RFPFoldXStabilityAndSASAProblemFactory,
)
from .rfp_rasp.register import RFPRaspBlackBox, RFPRaspProblemFactory
from .rmf_landscape.register import RMFBlackBox, RMFProblemFactory
from .rosetta_energy.register import RosettaEnergyBlackBox, RosettaEnergyProblemFactory
from .sa_tdc.register import SABlackBox, SAProblemFactory
from .scaffold_hop.register import ScaffoldHopBlackBox, ScaffoldHopProblemFactory
from .sitagliptin_mpo.register import (
    SitagliptinMPOBlackBox,
    SitagliptinMPOProblemFactory,
)
from .super_mario_bros.register import (
    SuperMarioBrosBlackBox,
    SuperMarioBrosProblemFactory,
)
from .thiothixene_rediscovery.register import (
    ThiothixeneRediscoveryBlackBox,
    ThiothixeneRediscoveryProblemFactory,
)
from .toy_continuous_problem.register import (
    ToyContinuousBlackBox,
    ToyContinuousProblemFactory,
)
from .troglitazone_rediscovery.register import (
    TroglitazoneRediscoveryBlackBox,
    TroglitazoneRediscoveryProblemFactory,
)
from .valsartan_smarts.register import (
    ValsartanSMARTSBlackBox,
    ValsartanSMARTSProblemFactory,
)
from .white_noise.register import WhiteNoiseBlackBox, WhiteNoiseProblemFactory
from .zaleplon_mpo.register import ZaleplonMPOBlackBox, ZaleplonMPOProblemFactory

THIS_DIR = Path(__file__).parent.resolve()

# The objective repository is made of
# all the directories that are here:
AVAILABLE_OBJECTIVES = []
for d in THIS_DIR.glob("*"):
    if d.is_dir() and d.name != "__pycache__":
        AVAILABLE_OBJECTIVES.append(d.name)

        if (d / "isolated_function.py").exists():
            AVAILABLE_OBJECTIVES.append(f"{d.name}__isolated")

AVAILABLE_OBJECTIVES = sorted(AVAILABLE_OBJECTIVES)

# Available problem factories
# maps names of problems to their respective
# problem factories
AVAILABLE_PROBLEM_FACTORIES = {
    "aloha": AlohaProblemFactory,
    "ehrlich": EhrlichProblemFactory,
    "ehrlich_holo": EhrlichHoloProblemFactory,
    "dockstring": DockstringProblemFactory,
    "drd3_docking": DRD3ProblemFactory,
    "foldx_rfp_lambo": FoldXRFPLamboProblemFactory,
    "foldx_sasa": FoldXSASAProblemFactory,
    "foldx_stability": FoldXStabilityProblemFactory,
    "foldx_stability_and_sasa": FoldXStabilityAndSASAProblemFactory,
    "gfp_cbas": GFPCBasProblemFactory,
    "gfp_select": GFPSelectionProblemFactory,
    "penalized_logp_lambo": PenalizedLogPLamboProblemFactory,
    "rasp": RaspProblemFactory,
    "rdkit_logp": LogPProblemFactory,
    "rdkit_qed": QEDProblemFactory,
    "rfp_rasp": RFPRaspProblemFactory,
    "rfp_foldx_stability": RFPFoldXStabilityProblemFactory,
    "rfp_foldx_stability_and_sasa": RFPFoldXStabilityAndSASAProblemFactory,
    "rmf_landscape": RMFProblemFactory,
    "rosetta_energy": RosettaEnergyProblemFactory,
    "sa_tdc": SAProblemFactory,
    "super_mario_bros": SuperMarioBrosProblemFactory,
    "white_noise": WhiteNoiseProblemFactory,
    "toy_continuous_problem": ToyContinuousProblemFactory,
    "gsk3_beta": GSK3BetaProblemFactory,
    "drd2_docking": DRD2ProblemFactory,
    "jnk3": JNK3ProblemFactory,
    "celecoxib_rediscovery": CelecoxibRediscoveryProblemFactory,
    "thiothixene_rediscovery": ThiothixeneRediscoveryProblemFactory,
    "troglitazone_rediscovery": TroglitazoneRediscoveryProblemFactory,
    "albuterol_similarity": AlbuterolSimilarityProblemFactory,
    "mestranol_similarity": MestranolSimilarityProblemFactory,
    "amlodipine_mpo": AmlodipineMPOProblemFactory,
    "fexofenadine_mpo": FexofenadineMPOProblemFactory,
    "osimetrinib_mpo": OsimetrinibMPOProblemFactory,
    "perindopril_mpo": PerindoprilMPOProblemFactory,
    "ranolazine_mpo": RanolazineMPOProblemFactory,
    "sitagliptin_mpo": SitagliptinMPOProblemFactory,
    "zaleplon_mpo": ZaleplonMPOProblemFactory,
    "deco_hop": DecoHopProblemFactory,
    "scaffold_hop": ScaffoldHopProblemFactory,
    "isomer_c7h8n2o2": IsomerC7H8N2O2ProblemFactory,
    "isomer_c9h10n2o2pf2cl": IsomerC9H10N2O2PF2ClProblemFactory,
    "median_1": Median1ProblemFactory,
    "median_2": Median2ProblemFactory,
    "valsartan_smarts": ValsartanSMARTSProblemFactory,
}

AVAILABLE_BLACK_BOXES = {
    "aloha": AlohaBlackBox,
    "ehrlich": EhrlichBlackBox,
    "ehrlich_holo": EhrlichHoloBlackBox,
    "dockstring": DockstringBlackBox,
    "drd3_docking": DRD3BlackBox,
    "foldx_rfp_lambo": FoldXRFPLamboBlackBox,
    "foldx_sasa": FoldXSASABlackBox,
    "foldx_stability": FoldXStabilityBlackBox,
    "foldx_stability_and_sasa": FoldXStabilityAndSASABlackBox,
    "gfp_cbas": GFPCBasBlackBox,
    "gfp_select": GFPSelectionBlackBox,
    "penalized_logp_lambo": PenalizedLogPLamboBlackBox,
    "rasp": RaspBlackBox,
    "rdkit_logp": LogPBlackBox,
    "rdkit_qed": QEDBlackBox,
    "rfp_rasp": RFPRaspBlackBox,
    "rfp_foldx_stability": RFPFoldXStabilityBlackBox,
    "rfp_foldx_stability_and_sasa": FoldXStabilityAndSASABlackBox,
    "rmf_landscape": RMFBlackBox,
    "rosetta_energy": RosettaEnergyBlackBox,
    "sa_tdc": SABlackBox,
    "super_mario_bros": SuperMarioBrosBlackBox,
    "white_noise": WhiteNoiseBlackBox,
    "toy_continuous_problem": ToyContinuousBlackBox,
    "gsk3_beta": GSK3BetaBlackBox,
    "drd2_docking": DRD2BlackBox,
    "jnk3": JNK3BlackBox,
    "celecoxib_rediscovery": CelecoxibRediscoveryBlackBox,
    "thiothixene_rediscovery": ThiothixeneRediscoveryBlackBox,
    "troglitazone_rediscovery": TroglitazoneRediscoveryBlackBox,
    "albuterol_similarity": AlbuterolSimilarityBlackBox,
    "mestranol_similarity": MestranolSimilarityBlackBox,
    "amlodipine_mpo": AmlodipineMPOBlackBox,
    "fexofenadine_mpo": FexofenadineMPOBlackBox,
    "osimetrinib_mpo": OsimetrinibMPOBlackBox,
    "perindopril_mpo": PerindoprilMPOBlackBox,
    "ranolazine_mpo": RanolazineMPOBlackBox,
    "sitagliptin_mpo": SitagliptinMPOBlackBox,
    "zaleplon_mpo": ZaleplonMPOBlackBox,
    "deco_hop": DecoHopBlackBox,
    "scaffold_hop": ScaffoldHopBlackBox,
    "isomer_c7h8n2o2": IsomerC7H8N2O2BlackBox,
    "isomer_c9h10n2o2pf2cl": IsomerC9H10N2O2PF2ClBlackBox,
    "median_1": Median1BlackBox,
    "median_2": Median2BlackBox,
    "valsartan_smarts": ValsartanSMARTSBlackBox,
}


def get_problems():
    return sorted(list(AVAILABLE_PROBLEM_FACTORIES.keys()))



import pytest


@pytest.mark.poli__ehrlich_holo
def test_ehrlich_holo_factory():
    from poli.objective_repository import EhrlichHoloProblemFactory

    problem = EhrlichHoloProblemFactory().create(
        sequence_length=10,
        motif_length=3,
        n_motifs=2,
        epistasis_factor=0.5,
    )
    f, x0 = problem.black_box, problem.x0
    print(f(x0))


@pytest.mark.poli__ehrlich_holo
def test_ehrlich_holo_builds_and_queries():
    from poli.objective_repository import EhrlichHoloBlackBox

    black_box = EhrlichHoloBlackBox(
        sequence_length=10,
        motif_length=3,
        n_motifs=2,
        epistasis_factor=0.5,
    )
    x0 = black_box.initial_solution()
    print(black_box(x0))

    x_final = black_box.optimal_solution()
    print(black_box(x_final))


@pytest.mark.poli__ehrlich_holo
def test_ehrlich_holo_works_on_isolation():
    from poli.objective_repository import EhrlichHoloBlackBox

    black_box = EhrlichHoloBlackBox(
        sequence_length=10,
        motif_length=3,
        n_motifs=2,
        epistasis_factor=0.0,
        force_isolation=True,
    )
    x0 = black_box.initial_solution(n_samples=100)
    print(black_box(x0))

    x_final = black_box.optimal_solution()
    print(black_box(x_final))


@pytest.mark.poli__ehrlich_holo
def test_ehrlich_seed_determinism():
    from poli.objective_repository import EhrlichHoloBlackBox

    black_box = EhrlichHoloBlackBox(
        sequence_length=10,
        motif_length=3,
        n_motifs=2,
        epistasis_factor=0.0,
        seed=42,
    )
    x0 = black_box.initial_solution()
    print(black_box(x0))

    black_box_2 = EhrlichHoloBlackBox(
        sequence_length=10,
        motif_length=3,
        n_motifs=2,
        epistasis_factor=0.0,
        seed=42,
    )
    x0_2 = black_box.initial_solution()
    print(black_box_2(x0_2))

    assert (black_box(x0) == black_box_2(x0_2)).all()
    assert (x0 == x0_2).all()



"""Script that gets called by the mother process to start an external observer process.
"""

import argparse
import sys
import traceback

from poli.core.util.abstract_observer import AbstractObserver
from poli.core.util.inter_process_communication.process_wrapper import get_connection
from poli.external_isolated_function_script import dynamically_instantiate


def start_observer_process(observer_name, port: int, password: str):
    """
    Starts the observer process.

    Parameters
    ----------
    observer_name : str
        The name of the observer to instantiate.
    port : int
        The port number for the connection with the mother process.
    password : str
        The password for the connection with the mother process.

    Notes
    -----
    This function starts the observer process by establishing a connection with the mother process,
    receiving setup information, instantiating the observer, initializing the observer with the setup
    information, and then waiting for observe calls. If an exception occurs during observation or
    attribute retrieval, it is sent back to the mother process.

    The observer process can be terminated by sending a "QUIT" message.
    """
    # make connection with the mother process
    conn = get_connection(port, password)

    # get setup info from external_observer
    setup_info, caller_info, seed, observer_kwargs = conn.recv()

    # instantiate observer
    observer: AbstractObserver = dynamically_instantiate(
        observer_name, **observer_kwargs
    )

    try:
        observer_info = observer.initialize_observer(setup_info, caller_info, seed)
        # give mother process the signal that we're ready
        conn.send(["SETUP", observer_info])
    except Exception as e:
        tb = traceback.format_exc()
        conn.send(["EXCEPTION", e, tb])
        sys.exit(1)

    # now wait for observe calls
    while True:
        msg_type, *msg = conn.recv()
        if msg_type == "OBSERVATION":
            # How should we inform the external observer
            # if something fails during observation?
            # (TODO).
            try:
                observer.observe(*msg)
                conn.send(["OBSERVATION", None])
            except Exception as e:
                tb = traceback.format_exc()
                conn.send(["EXCEPTION", e, tb])
        elif msg_type == "LOG":
            # How should we inform the external observer
            # if something fails during observation?
            # (TODO).
            try:
                observer.log(*msg)
                conn.send(["LOG", None])
            except Exception as e:
                tb = traceback.format_exc()
                conn.send(["EXCEPTION", e, tb])
        elif msg_type == "ATTRIBUTE":
            try:
                conn.send(["ATTRIBUTE", getattr(observer, msg[0])])
            except Exception as e:
                tb = traceback.format_exc()
                conn.send(["EXCEPTION", e, tb])
        elif msg_type == "QUIT":
            break
    observer.finish()
    # conn.close()
    # exit()  # kill other threads, and close file handles


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--objective-name", required=True)
    parser.add_argument("--port", required=True, type=int)
    parser.add_argument("--password", required=True, type=str)

    args, _ = parser.parse_known_args()
    start_observer_process(args.objective_name, args.port, args.password)



"""Implement a wrapper around the Therapeutics Data Commons (TDC) oracles [1].

When run, this script registers the TDCIsolatedFunction class as an
isolated function with the name "tdc__isolated".

References
----------
[1] “Artificial Intelligence Foundation for Therapeutic Science.”
    Huang, Kexin, Tianfan Fu, Wenhao Gao, Yue Zhao, Yusuf Roohani, Jure Leskovec,
    Connor W. Coley, Cao Xiao, Jimeng Sun, and Marinka Zitnik.  Nature Chemical Biology 18, no. 10
    (October 2022): 1033-36. https://doi.org/10.1038/s41589-022-01131-2.
"""

import numpy as np
from tdc import Oracle

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.util.chemistry.string_to_molecule import translate_selfies_to_smiles


class TDCIsolatedFunction(AbstractIsolatedFunction):
    """
    TDCIsolatedFunction is a class that represents the isolated logic
    for the TDC (Therapeutics Data Commons) problems.

    Parameters
    -----------
    oracle_name : str
        The name of the oracle used for computing the docking score.
    from_smiles : bool, optional
        Flag indicating whether the input molecules are in SMILES format. Defaults to True.
    **kwargs_for_oracle : dict, optional
        Additional keyword arguments passed to the oracle.

    Attributes
    ----------
    oracle : Oracle
        An instance of the Oracle class from TDC. It is instantiated as
        Oracle(name=oracle_name, **kwargs_for_oracle)
    from_smiles : bool
        Flag indicating whether the input molecules are in SMILES format.
        False indicates that the input molecules are in SELFIES format.
    """

    def __init__(
        self,
        oracle_name: str,
        from_smiles: bool = True,
        **kwargs_for_oracle,
    ):
        """
        Initialize the TDCIsolatedFunction class.

        Parameters
        ----------
        oracle_name : str
            The name of the oracle.
        from_smiles : bool, optional
            Whether to use SMILES representation, by default True.
        **kwargs_for_oracle : dict, optional
            Additional keyword arguments for the oracle.
        """
        super().__init__()
        self.oracle = Oracle(name=oracle_name, **kwargs_for_oracle)
        self.from_smiles = from_smiles

    def __call__(self, x, context=None):
        """
        Assuming x is an array of strings,
        we concatenate them and then
        compute the oracle score.

        Parameters
        -----------
        x : array-like
            An array of strings representing the input molecules.
        context : any, optional
            Additional context information. Defaults to None.

        Returns
        --------
        scores : array-like
            An array of oracle scores computed for each input molecule.
        """
        if x.dtype.kind not in ["U", "S"]:
            raise ValueError(
                f"We expect x to be an array of strings, but we got {x.dtype}"
            )

        molecule_strings = ["".join([x_ij for x_ij in x_i.flatten()]) for x_i in x]

        if not self.from_smiles:
            molecule_strings = translate_selfies_to_smiles(molecule_strings)

        scores = []
        for molecule_string in molecule_strings:
            scores.append(self.oracle(molecule_string))

        return np.array(scores).reshape(-1, 1)


if __name__ == "__main__":
    from poli.core.registry import register_isolated_function

    register_isolated_function(
        TDCIsolatedFunction,
        name="tdc__isolated",
        conda_environment_name="poli__tdc",
    )



from __future__ import annotations

import logging
import os
from pathlib import Path

import numpy as np
import pyrosetta
from pyrosetta.rosetta.core.pack.task import TaskFactory
from pyrosetta.rosetta.core.pose import deep_copy
from pyrosetta.rosetta.core.scoring import (
    ScoreFunctionFactory,
    fa_rep,
    get_score_function,
)
from pyrosetta.rosetta.protocols.loops import get_fa_scorefxn
from pyrosetta.rosetta.protocols.minimization_packing import MinMover, PackRotamersMover
from pyrosetta.rosetta.protocols.relax import FastRelax
from pyrosetta.toolbox import cleanATOM, mutate_residue

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.util.proteins.pdb_parsing import parse_pdb_as_residue_strings

pyrosetta.init()

ROSETTA_ENERGY_DIR = Path(__file__).parent.resolve()


class RosettaEnergyIsolatedLogic(AbstractIsolatedFunction):
    """
    RosettaEnergy internal implementation.

    Parameters
    ----------
    wildtype_pdb_path : Path
        The path to the wildtype PDB file(s), by default None.
        NOTE: currently only for single PDB files implemented -- List of Paths not yet supported!
    score_function : str, optional
        Which Rosetta score function to use. Options are [ref2015 , default , centroid , fa].
        The default function references ref2015.
    seeds : int, optional
        Overwrite Rosetta random seed with own integer, uses mt19937 RT reference (as per Rosetta default).
    unit : str, optional
        Output unit of black-box. Default is delta REU, which is difference between variant and wild-type.
        Alternatives are:
            REU -- raw energy function computed value,
            DDG -- scaled delta REU
    conversion_factor : float, optional
        Scaling factor applied when computing DDGs (DDG=deltaREU/conversion_factor). Only applies to DDG conversion.
        Default is '2.9' following:
            Park, Hahnbeom, et al.
            "Simultaneous optimization of biomolecular energy functions on features from small molecules and macromolecules."
            Journal of chemical theory and computation 12.12 (2016): 6201-6212.
            DOI: https://doi.org/10.1021/acs.jctc.6b00819
    clean : bool, optional
        Flag whether to apply PDB cleaning routine prior to pose loading of reference.
        Default is True.
    relax: bool, optional
        Flag whether to apply relaxation protocol. Default enabled.
        NOTE: disable if fast compute required
    pack: bool, optional
        Flag whether to apply packing protocol. Default enabled.
        NOTE: disable if fast compute required
    cycle: int, optional
        Number of relaxation cycles applied.
    constraint_weight: int, optional
        Constraint on scoring function for atom-pair, angles, coordinates.

    Methods
    -------
    _apply_on_sequence(x, context=None)
        The main black box method that performs the computation, i.e.
        it computes the energy units of the mutant(s) in x.
    _check_pose_steric_clashes(mutant_pose)
        Checks for steric clashes of variant pose against WT pose, returns boolean.
    _apply_relax_pack(mutant_pose)
        Apply pack and relax routines on pose.

    Raises
    ------
    AssertionError
        If wildtype_pdb_path is not provided.
        If the provided energy units are not valid.
    NotImplementedError
        If energy function does not exist.

    Notes
    -----
    - The wildtype_pdb_path is a single Path object.
    """

    def __init__(
        self,
        wildtype_pdb_path: Path,
        score_function: str = "default",
        seed: int = 0,
        unit: str = "DDG",
        conversion_factor: float = 2.9,
        clean: bool = True,
        relax: bool = True,
        pack: bool = True,
        cycle: int = 3,
        constraint_weight: float | int = 5,
        n_threads: int = 4,
    ):
        # set number of threads prior to initializing Rosetta
        os.environ["OMP_NUM_THREADS"] = str(n_threads)
        pyrosetta.init(extra_options=f"-nthreads {n_threads}")
        self.conversion_factor = conversion_factor  # iff ddGs computed from REUs
        self.relax = relax  # do we use FastRelax?
        self.cycle = cycle  # iterations on relax protocol
        self.constraint_weight = constraint_weight
        valid_units = ["DDG", "REU", "DREU"]
        if unit.upper() not in valid_units:
            raise AssertionError(f"Output unit {unit} not a valid unit: {valid_units}")
        self.unit = unit.upper()
        if seed is not None:
            pyrosetta.rosetta.basic.random.init_random_generators(seed, "mt19937")
        self.score_function_identifier = score_function
        self.energy_function = self.__get_score_fn(score_function)
        self.energy_function.set_weight(
            pyrosetta.rosetta.core.scoring.ScoreType.atom_pair_constraint,
            self.constraint_weight,
        )
        self.energy_function.set_weight(
            pyrosetta.rosetta.core.scoring.ScoreType.coordinate_constraint,
            self.constraint_weight,
        )
        self.energy_function.set_weight(
            pyrosetta.rosetta.core.scoring.ScoreType.angle_constraint,
            self.constraint_weight,
        )
        self.x0 = np.array([parse_pdb_as_residue_strings(wildtype_pdb_path)])
        if not isinstance(wildtype_pdb_path, Path):
            raise TypeError(f"{wildtype_pdb_path} is not a Path")
        if clean:
            cleanATOM(wildtype_pdb_path)
            self.pose = pyrosetta.pose_from_pdb(
                str(wildtype_pdb_path.with_suffix(".clean.pdb"))
            )
        else:
            self.pose = pyrosetta.pose_from_pdb(str(wildtype_pdb_path))
        # set packing protocol
        self.pack = None
        if pack:
            pack_task = TaskFactory.create_packer_task(self.pose)
            pack_task.restrict_to_repacking()
            self.packer = PackRotamersMover(self.energy_function, pack_task)
            self.packer.apply(self.pose)
        self.relax = None
        # set MinMover; NOTE: suitable for small changes to structure
        self.mover = MinMover()
        # self.mover.min_type("lbfgs_armijo_nonmonotone")
        self.mover.score_function(self.energy_function)
        # set relax protocol; NOTE: relax & pack required when larger/destabilizing changes are introduced
        if relax:
            self.relax = FastRelax(self.energy_function, self.cycle)
            cartesian_scoring = bool("cart" in self.score_function_identifier)
            self.relax.cartesian(cartesian_scoring)
            self.relax.minimize_bond_angles(cartesian_scoring)
            self.relax.minimize_bond_lengths(cartesian_scoring)
            self.relax.apply(self.pose)
        self.wt_score = self.energy_function.score(self.pose)
        self.wt_fa_rep = self.pose.energies().total_energies()[fa_rep]
        self.x_t = None  # track sequences as property

    def __get_score_fn(self, score_function_identifier: str):
        if score_function_identifier == "default":
            return get_score_function(True)
        elif score_function_identifier == "ref2015":
            return get_score_function(True)
        elif score_function_identifier == "ref2015_cart":
            return ScoreFunctionFactory.create_score_function(score_function_identifier)
        elif score_function_identifier == "centroid":
            return get_score_function(False)
        elif score_function_identifier == "franklin2019":
            # TODO: requires AddMembraneMover , not yet functional!
            return ScoreFunctionFactory.create_score_function(score_function_identifier)
        elif score_function_identifier == "fa":
            return get_fa_scorefxn()
        else:
            raise NotImplementedError("Invalid scoring function!")

    def __call__(self, x: np.ndarray, context=None) -> np.ndarray:
        self.x_t = []
        y = np.empty([x.shape[0], 1])
        for i in range(x.shape[0]):
            y[i, 0] = self._apply_on_sequence(x[i, :])
        return y

    def _apply_on_sequence(self, seq_arr: str | np.ndarray) -> float:
        if not isinstance(seq_arr, np.ndarray):
            seq_arr = np.array(list(seq_arr))
        seq_arr = np.atleast_2d(seq_arr)
        pose_mutant = deep_copy(self.pose)
        diff_rosetta_idx = (
            np.where(self.x0 != seq_arr)[1] + 1
        )  # get diff indices (Rosetta indexes at 1)
        diff_residues = seq_arr[self.x0 != seq_arr]  # get residue mutant diff to WT
        for idx, res in zip(diff_rosetta_idx, diff_residues):
            mutate_residue(pose_mutant, idx, res)  # inplace mutation on pose copy
        self.x_t.append(pose_mutant.sequence())
        self.mover.apply(pose_mutant)
        REU = self.energy_function.score(pose_mutant)
        clashes = False
        if self.relax is not None and self.pack is not None:
            clashes = self._check_pose_steric_clashes(mutant_pose=pose_mutant)
        if clashes:
            logging.info(
                "Steric clashes detected, attempt recovery by relax and pack-minimize..."
            )
            self._apply_relax_pack(pose_mutant)
            logging.info("Recompute energy function")
            REU = self.energy_function.score(pose_mutant)
        if self.unit == "REU":
            return REU
        elif self.unit == "DREU":
            return REU - self.wt_score
        elif self.unit == "DDG":
            return (REU - self.wt_score) / self.conversion_factor
        else:
            RuntimeError("Output unit not specified correctly!")

    def _check_pose_steric_clashes(self, mutant_pose: object) -> bool:
        mutant_fa_rep = mutant_pose.energies().total_energies()[fa_rep]
        return bool(mutant_fa_rep > self.wt_fa_rep)

    def _apply_relax_pack(self, mutant_pose: object) -> None:
        self.pack.apply(mutant_pose)
        self.relax.apply(mutant_pose)


if __name__ == "__main__":
    from poli.core.registry import register_isolated_function

    register_isolated_function(
        RosettaEnergyIsolatedLogic,
        name="rosetta_energy__isolated",
        conda_environment_name="poli__rosetta_energy",
        force=True,
    )



"""Utilities for transforming string representations of
molecules into molecules in RDKit.
"""

from typing import List

import selfies as sf
from rdkit import Chem


def translate_smiles_to_selfies(
    smiles_strings: List[str], strict: bool = False
) -> List[str]:
    """Translates a list of SMILES strings to SELFIES strings.

    Given a list of SMILES strings, returns the translation
    into SELFIES strings. If strict is True, it raises an error
    if a SMILES string in the list cannot be parsed. Else, it
    returns None for those.

    This function uses the `selfies` package from Aspuru-Guzik's
    lab. See https://github.com/aspuru-guzik-group/selfies


    Parameters
    ----------
    smiles_strings : List[str]
        A list of SMILES strings.
    strict : bool, optional
        If True, raise an error if a SMILES string in the list cannot be parsed.

    Returns
    -------
    List[str]
        A list of SELFIES strings.
    """
    selfies_strings = []
    for smile in smiles_strings:
        try:
            selfies_strings.append(sf.encoder(smile))
        except sf.EncoderError:
            if strict:
                raise ValueError("Failed to encode SMILES to SELFIES.")
            else:
                selfies_strings.append(None)

    return selfies_strings


def translate_selfies_to_smiles(
    selfies_strings: List[str], strict: bool = False
) -> List[str]:
    """Translates a list of SELFIES strings to SMILES strings.

    Given a list of SELFIES strings, returns the translation
    into SMILES strings. If strict is True, it raises an error
    if a SELFIES string in the list cannot be parsed. Else, it
    returns None for those.

    This function uses the `selfies` package from Aspuru-Guzik's
    lab. See https://github.com/aspuru-guzik-group/selfies

    Parameters
    ----------
    selfies_strings : List[str]
        A list of SELFIES strings.
    strict : bool, optional
        If True, raise an error if a SELFIES string in the list cannot be parsed.

    Returns
    -------
    smiles_strings : List[str]
        A list of SMILES strings.
    """
    smiles_strings = []
    for selfies in selfies_strings:
        try:
            smiles_strings.append(sf.decoder(selfies))
        except sf.DecoderError:
            if strict:
                raise ValueError("Failed to decode SELFIES to SMILES.")
            else:
                smiles_strings.append(None)

    return smiles_strings


def smiles_to_molecules(smiles_strings: List[str], strict: bool = False) -> Chem.Mol:
    """Converts a list of SMILES strings to RDKit molecules.

    Converts a list of SMILES strings to RDKit molecules. If strict is True,
    it raises an error if a SMILES string in the list cannot be parsed.

    Parameters
    ----------
    smiles : List[str]
        A list of SMILES string.

    Returns
    -------
    Chem.Mol
        An RDKit molecule.
    """
    molecules = []
    for smiles_string in smiles_strings:
        molecule = Chem.MolFromSmiles(smiles_string)

        if molecule is None and strict:
            raise ValueError(f"Failed to parse molecule {smiles_string}.")

        molecules.append(molecule)

    return molecules


def selfies_to_molecules(selfies_strings: List[str]) -> Chem.Mol:
    """Converts a list of selfies strings to RDKit molecules.

    Parameters
    ----------
    selfies : str
        A selfies string.
    strict : bool, optional
        If True, raise an error if selfies fails to decode the selfies string.

    Returns
    -------
    Chem.Mol
        An RDKit molecule.
    """
    smiles_strings = translate_selfies_to_smiles(selfies_strings, strict=True)
    molecule = smiles_to_molecules(smiles_strings)
    return molecule


def strings_to_molecules(molecule_strings: List[str], from_selfies: bool = False):
    """
    Convert a string representation of a molecule to an RDKit molecule.

    Parameters
    ----------
    molecule_string : str
        A string representation of a molecule.
    from_selfies : bool, optional
        If True, the molecule string is assumed to be a selfies string.
        Otherwise, the molecule string is assumed to be a SMILES string.

    Returns
    -------
    Chem.Mol
        An RDKit molecule.
    """
    if from_selfies:
        return selfies_to_molecules(molecule_strings)
    else:
        return smiles_to_molecules(molecule_strings)



"""
A series of testbed functions for optimization.

See for more examples:
https://en.wikipedia.org/wiki/Test_functions_for_optimization
"""

from typing import List, Literal

import numpy as np

from .definitions import (
    ackley_function_01,
    alpine_01,
    alpine_02,
    bent_cigar,
    branin_2d,
    brown,
    camelback_2d,
    chung_reynolds,
    cosine_mixture,
    cross_in_tray,
    deb_01,
    deb_02,
    deflected_corrugated_spring,
    easom,
    egg_holder,
    hartmann_6d,
    levy,
    rosenbrock,
    shifted_sphere,
    styblinski_tang,
)

# Notice: these will be used by pytest to test the
# instancing of the toy objective functions.
POSSIBLE_FUNCTIONS = [
    "ackley_function_01",
    "alpine_01",
    "alpine_02",
    "bent_cigar",
    "brown",
    "chung_reynolds",
    "cosine_mixture",
    "deb_01",
    "deb_02",
    "deflected_corrugated_spring",
    "styblinski_tang",
    "shifted_sphere",
    "easom",
    "cross_in_tray",
    "egg_holder",
    "camelback_2d",
    "hartmann_6d",
    "branin_2d",
    "rosenbrock",
    "levy",
]
TWO_DIMENSIONAL_PROBLEMS = [
    "easom",
    "cross_in_tray",
    "egg_holder",
    "camelback_2d",
    "branin_2d",
]
SIX_DIMENSIONAL_PROBLEMS = ["hartmann_6d"]


class ToyContinuousProblem:
    """
    Contains the toy objective functions, their limits, and the optima location.

    For more information, check definitions.py and [1].

    [1]:  https://al-roomi.org/benchmarks/unconstrained/n-dimensions
    """

    def __init__(
        self,
        name: Literal[
            "ackley_function_01",
            "alpine_01",
            "alpine_02",
            "bent_cigar",
            "brown",
            "chung_reynolds",
            "cosine_mixture",
            "deb_01",
            "deb_02",
            "deflected_corrugated_spring",
            "styblinski_tang",
            "shifted_sphere",
            "easom",
            "cross_in_tray",
            "egg_holder",
            "camelback_2d",
            "branin_2d",
            "hartmann_6d",
            "rosenbrock",
            "levy",
        ],
        n_dims: int = 2,
        embed_in: int = None,
        dimensions_to_embed_in: List[int] = None,
    ) -> None:
        self.maximize = True
        self.known_optima = True
        self.dimensions_to_embed_in = dimensions_to_embed_in

        if n_dims != 2 and name in TWO_DIMENSIONAL_PROBLEMS:
            if embed_in is None:
                raise ValueError(
                    f"Function {name} can only be instantiated in two dimensions (received {n_dims})."
                    " Alternatively, you can embed the function in higher dimensions by setting"
                    " embed_in: int to the desired dimension. When doing so, the 2 dimensions will be "
                    "randomly selected among the embed_in."
                )

        if dimensions_to_embed_in is not None:
            assert (
                embed_in is not None
            ), "Expected dimensions_to_embed_in to be None if embed_in is None."
            for dim in dimensions_to_embed_in:
                if dim >= embed_in or dim < 0:
                    raise ValueError(
                        f"Dimension to embed in {dim} is higher than the number of dimensions {n_dims} or negative."
                    )

        if name == "ackley_function_01":
            self.function = ackley_function_01
            self.limits = [-32.0, 32.0]
            self.optima_location = np.array([0.0] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[15.0] * self.solution_length])
        elif name == "alpine_01":
            self.function = alpine_01
            self.limits = [-10.0, 10.0]
            self.optima_location = np.array([0.0] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[8.0] * self.solution_length])
        elif name == "alpine_02":
            self.function = alpine_02
            self.limits = [0.0, 10.0]
            self.optima_location = np.array([7.9170526982459462172] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[3.0] * self.solution_length])
        elif name == "bent_cigar":
            self.function = bent_cigar
            self.limits = [-100.0, 100.0]
            self.optima_location = np.array([0.0] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[50.0] * self.solution_length])
        elif name == "brown":
            self.function = brown
            self.limits = [-1.0, 4.0]
            self.optima_location = np.array([0.0] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[3.0] * self.solution_length])
        elif name == "chung_reynolds":
            self.function = chung_reynolds
            self.limits = [-100.0, 100.0]
            self.optima_location = np.array([0.0] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[50.0] * self.solution_length])
        elif name == "cosine_mixture":
            self.function = cosine_mixture
            self.limits = [-1.0, 1.0]
            self.optima_location = np.array([0.0] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[-0.75] * self.solution_length])
        elif name == "deb_01":
            self.function = deb_01
            self.limits = [-1.0, 1.0]
            self.optima_location = np.array([0.1] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[-0.75] * self.solution_length])
        elif name == "deb_02":
            self.function = deb_02
            self.limits = [0.0, 1.0]
            self.optima_location = np.array([0.105 ** (4 / 3)] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[0.75] * self.solution_length])
        elif name == "deflected_corrugated_spring":
            self.function = deflected_corrugated_spring
            self.limits = [0.0, 10.0]
            self.optima_location = np.array([5.0] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[1.0] * self.solution_length])
        elif name == "styblinski_tang":
            self.function = styblinski_tang
            self.limits = [-5.0, 5.0]
            self.optima_location = np.array([-2.903534] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[4.0] * self.solution_length])
        elif name == "shifted_sphere":
            self.function = shifted_sphere
            self.limits = [-4.0, 4.0]
            self.optima_location = np.array([1.0] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[-3.0] * self.solution_length])
        elif name == "easom":
            self.function = easom
            self.limits = [np.pi - 4, np.pi + 4]
            self.optima_location = np.array([np.pi, np.pi])
            self.solution_length = 2
            self.x0 = np.array([[0.0] * self.solution_length])
        elif name == "cross_in_tray":
            self.function = cross_in_tray
            self.limits = [-10.0, 10.0]
            self.optima_location = np.array([1.34941, 1.34941])
            self.solution_length = 2
            self.x0 = np.array([[-7.0] * self.solution_length])
        elif name == "egg_holder":
            self.function = egg_holder
            self.limits = [-700.0, 700.0]
            self.optima_location = np.array([512, 404.2319])
            self.solution_length = 2
            self.x0 = np.array([[-600.0] * self.solution_length])
        elif name == "camelback_2d":
            self.function = camelback_2d
            self.limits = [-5.0, 5.0]
            self.optima_location = np.array([0.0898, -0.7126])
            self.solution_length = 2
            self.x0 = np.array([[-3.0] * self.solution_length])
        elif name == "hartmann_6d":
            self.function = hartmann_6d
            self.limits = [0.0, 1.0]
            self.optima_location = np.array(
                [0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573]
            )
            self.solution_length = 6
            self.x0 = np.array([[0.8] * self.solution_length])
        elif name == "branin_2d":
            self.function = branin_2d
            self.limits = [-5.0, 15.0]
            self.optima_location = np.array([9.42478, 2.475])
            self.solution_length = 2
            self.x0 = np.array([0.0] * n_dims).reshape(1, n_dims)
        elif name == "rosenbrock":
            self.function = rosenbrock
            self.limits = [-5.0, 10.0]
            self.optima_location = np.array([1.0] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[0.0] * self.solution_length])
        elif name == "levy":
            self.function = levy
            self.limits = [-10.0, 10.0]
            self.optima_location = np.array([1.0] * n_dims)
            self.solution_length = n_dims
            self.x0 = np.array([[0.0] * self.solution_length])
        else:
            raise ValueError(f"Expected {name} to be one of {POSSIBLE_FUNCTIONS}")

        self.optima = self.function(self.optima_location.reshape(1, -1))

        # If embed_in is not None, then we will embed the
        # function in embed_in dimensions. This is useful for testing
        # algorithms that leverage low intrinsic dimensionality.
        # At this point, solution length is the intrinsic dimensionality
        # of the problem.
        if embed_in is not None:
            assert self.solution_length < embed_in, (
                f"Expected the intrinsic dimensionality of the problem to be lower than the "
                f"dimensionality of the space, but got {self.solution_length} and {embed_in} respectively."
            )

            if dimensions_to_embed_in is None:
                self.dimensions_to_embed_in = np.random.permutation(embed_in)[
                    : self.solution_length
                ]

            # We update the solution length to the embedded dimensionality.
            self.solution_length = embed_in
            previous_optima_location = self.optima_location.copy()
            previous_x0 = self.x0.copy()
            self.optima_location = np.zeros(embed_in)
            self.optima_location[self.dimensions_to_embed_in] = previous_optima_location
            self.x0 = np.zeros(embed_in)
            self.x0[self.dimensions_to_embed_in] = previous_x0.flatten()
            self.x0 = self.x0.reshape(1, -1)

            _current_function = self.function
            self.function = lambda x: _current_function(
                x[:, self.dimensions_to_embed_in]
            )
        else:
            # We need to make sure that the user specified a feasible
            # number of dimensions. If not, we raise an error.
            assert n_dims == self.solution_length, (
                f"The solution length for {name} should be {self.solution_length},"
                f" but received {n_dims}."
            )

    def evaluate_objective(self, x: np.array, **kwargs) -> np.array:
        return self.function(x)

    def __call__(self, x: np.array) -> np.array:
        return self.function(x).reshape(-1, 1)



"""
This script will send things to simulator.jar.

When ran, it lets a human play a level.
"""

import json
import subprocess
from pathlib import Path

import numpy as np
from level_utils import clean_level

filepath = Path(__file__).parent.resolve()
JARFILE_PATH = f"{filepath}/simulator.jar"


def test_level_from_int_array(
    level: np.ndarray,
    human_player: bool = False,
    max_time: int = 45,
    visualize: bool = False,
) -> dict:
    level = clean_level(level)
    level = str(level)

    return run_level(
        level, human_player=human_player, max_time=max_time, visualize=visualize
    )


def test_level_from_str_array(
    level: np.ndarray,
    human_player: bool = False,
    max_time: int = 45,
    visualize: bool = False,
) -> dict:
    level = str(level)

    return run_level(
        level, human_player=human_player, max_time=max_time, visualize=visualize
    )


def run_level(
    level: str,
    human_player: bool = False,
    max_time: int = 30,
    visualize: bool = False,
) -> dict:
    # Run the simulator.jar file with the given level
    if human_player:
        java = subprocess.Popen(
            ["java", "-cp", JARFILE_PATH, "geometry.PlayLevel", level],
            stdout=subprocess.PIPE,
        )
    else:
        java = subprocess.Popen(
            [
                "java",
                "-cp",
                JARFILE_PATH,
                "geometry.EvalLevel",
                level,
                str(max_time),
                str(visualize).lower(),
            ],
            stdout=subprocess.PIPE,
        )

    lines = java.stdout.readlines()
    res = lines[-1]
    res = json.loads(res.decode("utf8"))
    res["level"] = level

    return res



from typing import Literal, Union


class BenchmarkInformation:
    def __init__(
        self,
        name: str,
        max_sequence_length: int,
        aligned: bool,
        fixed_length: bool,
        deterministic: bool,
        alphabet: list,
        log_transform_recommended: bool = None,
        discrete: bool = True,
        fidelity: Union[Literal["high", "low"], None] = None,
        padding_token: str = "",
    ):
        self.name = name
        self.max_sequence_length = max_sequence_length
        self.aligned = aligned
        self.fixed_length = fixed_length
        self.deterministic = deterministic
        self.alphabet = alphabet
        self.log_transform_recommended = log_transform_recommended
        self.discrete = discrete
        self.fidelity = fidelity
        self.padding_token = padding_token

    def get_problem_name(self) -> str:
        """Returns the problem's name.

        Returns
        --------
        name : str
            The problem's name.
        """
        return self.name

    def get_max_sequence_length(self) -> int:
        """
        Returns the maximum sequence length allowed by the black-box.

        Returns
        --------
        max_sequence_length : int
            The length of the longest sequence.
        """
        return self.max_sequence_length

    def is_deterministic(self) -> bool:
        """
        Returns whether the black-box is deterministic.

        Returns
        --------
        deterministic : bool
            Whether the black-box is deterministic.
        """
        return self.deterministic

    def is_discrete(self) -> bool:
        """
        Returns whether the black-box has discrete inputs.

        Returns
        --------
        discrete : bool
            Whether the black-box has discrete inputs.
        """
        return self.discrete

    def get_padding_token(self) -> str:
        """
        Returns the padding token used by the black-box.

        By default, this is usually "", the empty string.

        Returns
        --------
        padding_token : str
            The padding token used by the black-box.
        """
        return self.padding_token

    def sequences_are_aligned(self) -> bool:
        """
        Returns whether the sequences need to be aligned.

        We defined aligned sequences as sequences that are aligned in such
        a way that the same position in each sequence corresponds to the
        same position in the other sequences.

        Problems can be aligned, but have several different lengths. One example
        is the protein sequence problem for several different wildtypes.

        Returns
        --------
        aligned : bool
            Whether the sequences need to be aligned.
        """
        return self.aligned

    def get_alphabet(self) -> list:
        """
        Returns the alphabet of allowed characters.

        Returns
        --------
        alphabet : list[str]
            List of tokens allowed by the black-box.
        """
        return self.alphabet

    def log_transform_recommended(self) -> bool:
        """
        Returns whether the black-box recommends log-transforming the targets.

        Returns
        --------
        log_transform_recommended : bool
            Whether the black-box recommends log-transforming the targets.
        """
        return self.log_transform_recommended

    def __str__(self):
        return f"BlackBoxInformation(name={self.name}, max_sequence_length={self.max_sequence_length}, aligned={self.aligned}, fixed_length={self.fixed_length}, discrete={self.discrete})"

    def __repr__(self):
        return f"<BlackBoxInformation(name={self.name}, max_sequence_length={self.max_sequence_length}, aligned={self.aligned}, fixed_length={self.fixed_length}, discrete={self.discrete}, alphabet={self.alphabet}, log_transform_recommended={self.log_transform_recommended})>"

    def as_dict(self):
        return {
            "name": self.name,
            "max_sequence_length": self.max_sequence_length,
            "aligned": self.aligned,
            "fixed_length": self.fixed_length,
            "deterministic": self.deterministic,
            "discrete": self.discrete,
            "fidelity": self.fidelity,
            "alphabet": self.alphabet,
            "log_transform_recommended": self.log_transform_recommended,
            "padding_token": self.padding_token,
        }



"""This module implements the abstract black box class, from which
all objective functions should inherit.
"""

from multiprocessing import Pool, cpu_count
from warnings import warn

import numpy as np

from poli.core.black_box_information import BlackBoxInformation
from poli.core.exceptions import BudgetExhaustedException
from poli.core.util.abstract_observer import AbstractObserver
from poli.core.util.alignment import is_aligned_input
from poli.core.util.batch import batched


class AbstractBlackBox:
    """
    Abstract base class for a black box optimization problem.

    Parameters
    ----------
    batch_size : int, optional
        The batch size for evaluating the black box function. Default is None.
    parallelize : bool, optional
        Flag indicating whether to evaluate the black box function in parallel.
        Default is False.
    num_workers : int, optional
        The number of workers to use for parallel evaluation. Default is None,
        which uses half of the available CPU cores.
    evaluation_budget : int, optional
        The maximum number of evaluations allowed for the black box function.
        Default is float("inf").

    Attributes
    ----------
    observer : AbstractObserver or None
        The observer object for recording observations during evaluation.
    observer_info : object or None
        The information given by the observer after initialization.
    parallelize : bool
        Flag indicating whether to evaluate the black box function in parallel.
    num_workers : int
        The number of workers to use for parallel evaluation.
    batch_size : int or None
        The batch size for evaluating the black box function.

    Methods
    -------
    set_observer(observer)
        Set the observer object for recording observations during evaluation.
    reset_evaluation_budget()
        Reset the evaluation budget by setting the number of evaluations made
        to 0.
    __call__(x, context=None)
        Evaluate the black box function for the given input.
    _black_box(x, context=None)
        Abstract method for evaluating the black box function.
    terminate()
        Terminate the black box optimization problem.
    __enter__()
        Enter the context manager.
    __exit__(exc_type, exc_val, exc_tb)
        Exit the context manager.
    __del__()
        Destructor for the black box optimization problem.
    __neg__()
        Create a new black box with the objective function as the negative
        of the original one.
    """

    def __init__(
        self,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        """
        Initialize the AbstractBlackBox object.

        Parameters
        ----------
        batch_size : int, optional
            The batch size for parallel execution, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize the execution, by default False.
        num_workers : int, optional
            The number of workers for parallel execution, by default we use half the available CPUs.
        evaluation_budget : int, optional
            The maximum number of evaluations allowed for the black box function, by default float("inf").
        """
        self.observer = None
        self.observer_info = None
        self.parallelize = parallelize
        self.evaluation_budget = evaluation_budget
        self.num_evaluations = 0
        self.force_isolation = force_isolation

        if num_workers is None:
            num_workers = cpu_count() // 2

        self.num_workers = num_workers

        self.batch_size = batch_size

    def get_black_box_info(self) -> BlackBoxInformation:
        raise NotImplementedError(
            "Black box information must be implemented in subclasses."
        )

    @property
    def info(self) -> BlackBoxInformation:
        return self.get_black_box_info()

    def set_evaluation_budget(self, evaluation_budget: int):
        """
        Set the evaluation budget for the black box function.

        Parameters
        ----------
        evaluation_budget : int
            The maximum number of evaluations allowed for the black box function.
        """
        self.evaluation_budget = evaluation_budget

    def set_observer(self, observer: AbstractObserver):
        """
        Set the observer object for recording observations during evaluation.

        Parameters
        ----------
        observer : AbstractObserver
            The observer object.
        """
        if not isinstance(observer, AbstractObserver):
            warn(
                f"poli 🧪: Observer is not an instance of AbstractObserver: {observer} "
                "Are you sure you want to set this observer?"
            )
        self.observer = observer

    def set_observer_info(self, observer_info: object):
        """
        Set the observer information after initialization.

        Parameters
        ----------
        observer_info : object
            The information given by the observer after initialization.
        """
        self.observer_info = observer_info

    def reset_evaluation_budget(self):
        """Resets the evaluation budget by setting the number of evaluations made to 0."""
        self.num_evaluations = 0

    def __call__(self, x: np.array, context=None):
        """Calls the black box function.

        The purpose of this function is to enforce that inputs are equal across
        problems. The inputs are usually assumed to be strings, and all
        objective functions in our repository will assume that the inputs are
        strings. Some also allow for integers (i.e. token ids) to be passed as
        inputs. Some problems have continuous inputs, too.

        Parameters
        ----------
        x : np.array
            The input to the black box function.
        context : object, optional
            Additional context information for the evaluation, by default None.

        Returns
        -------
        y: np.array
            The output of the black box function.

        Raises
        ------
        AssertionError
            If the input is not a 2D array (when the sequences were supposed to be aligned).
        AssertionError
            If the length of the input does not match the maximum sequence length of the problem.
        AssertionError
            If the output is not a 2D array.
        AssertionError
            If the length of the output does not match the length of the input.
        """
        # Maximum sequence length: This can be an integer L, or np.inf, or None.
        maximum_sequence_length = self.info.get_max_sequence_length()

        # We assume that the input is either a 2D array of shape [b, L],
        # or a 1D array of shape [b,]. If the input is a 1D array, then
        # we inflate the input to be [b, 1] (allowing for variable-length
        # inputs)
        if self.info.sequences_are_aligned():
            # If the user passed an array of shape [b,],
            # then we need to make sure that the sequences
            # are aligned. Otherwise, we will raise an error.
            assert is_aligned_input(
                x, maximum_sequence_length=maximum_sequence_length
            ), (
                "The input given is not suitable for aligned problems. "
                "This may happen because the input is not an array of "
                "strings, the input is not of shape [b, L], or the input "
                "is of shape [b, ], but the lengths of the strings are "
                "not the same. "
                " Aligned problems, by definition, have as input arrays"
                " of the same length."
            )
        else:
            # At this point, we know that the sequences are not aligned.
            # Which means that the user is likely to pass an array [b,]
            # instead of an array [b, L].
            # We will reshape the input to be [b, 1] if it is not already.
            # This is necessary for batching.
            if len(x.shape) == 1:
                x = x.reshape(-1, 1)

        # We assert that the length matches L if the maximum sequence length
        # specified in the problem is different from np.inf or None.
        if (
            maximum_sequence_length not in [np.inf, None]
            and self.info.sequences_are_aligned()
        ):
            if len(x.shape) == 1:
                assert len(set([len(s) for s in x])) == 1, (
                    "The sequences are not aligned. "
                    "If you want to allow for variable-length inputs, set "
                    "the maximum length of the problem to be L=np.inf or L=None. "
                    "(and the aligned flag to False)."
                )
                x = x.reshape(-1, 1)
            else:
                assert x.shape[1] == maximum_sequence_length, (
                    "The length of the input is not the same as the length of the input of the problem. "
                    f"(L={maximum_sequence_length}, x.shape[1]={x.shape[1]}). "
                    "If you want to allow for variable-length inputs, set L=np.inf or L=None."
                )

        # Evaluate f by batches. If batch_size is None, then we evaluate
        # the whole input at once.
        if self.batch_size is None:
            batch_size = x.shape[0]
        else:
            batch_size = self.batch_size
        f_evals = []

        # Check whether we have enough budget to evaluate the black box function
        # in the current batch.
        if self.num_evaluations + batch_size > self.evaluation_budget:
            raise BudgetExhaustedException(
                f"Exhausted the evaluation budget of {self.evaluation_budget} evaluations."
                f" (tried to evaluate {batch_size}, but we have already"
                f" evaluated {self.num_evaluations}/{self.evaluation_budget})."
            )

        # We evaluate x in batches.
        for x_batch_ in batched(x, batch_size):
            # We reshape the batch to be 2D, even if the batch size is 1.
            x_batch = (
                np.concatenate(x_batch_, axis=0).reshape(len(x_batch_), -1)
                if batch_size > 1
                else np.array(x_batch_)
            )

            # We evaluate the batch in parallel if the user wants to.
            if self.parallelize:
                with Pool(self.num_workers) as pool:
                    f_batch_ = pool.starmap(
                        self._black_box, [(x.reshape(1, -1), context) for x in x_batch]
                    )
                    f_batch = np.array(f_batch_).reshape(len(x_batch_), -1)
            else:  # iterative treatment
                f_batch = self._black_box(x_batch, context)

            assert (
                len(f_batch.shape) == 2
            ), f"len(f(x).shape)={len(f_batch.shape)}, expected 2"

            assert isinstance(
                f_batch, np.ndarray
            ), f"type(f)={type(f_batch)}, not np.ndarray"

            assert (
                x_batch.shape[0] == f_batch.shape[0]
            ), f"Inconsistent evaluation axis=0 x={x_batch.shape} != black_box y={f_batch.shape}"

            # We pass the information to the observer, if any.
            if self.observer is not None:
                # observer logic s.t. observations are individual - later aggregate w.r.t batch_size
                if x_batch.shape[0] > 1:
                    for i in range(x_batch.shape[0]):
                        _x = np.atleast_2d(x_batch[i])
                        _y = np.atleast_2d(f_batch[i])
                        self.observer.observe(_x, _y, context)
                else:
                    self.observer.observe(x_batch, f_batch, context)

            f_evals.append(f_batch)

            # We update the number of evaluations.
            self.num_evaluations += x_batch.shape[0]

        # Finally, we append the results of the batches.
        f = np.concatenate(f_evals, axis=0)
        return f

    def _black_box(self, x: np.ndarray, context=None) -> np.ndarray:
        """
        Abstract method for evaluating the black box function.

        Parameters
        ----------
        x : np.ndarray
            The input values to evaluate the black box function.
        context : any, optional
            Additional context information for the evaluation. Defaults to None.

        Returns
        -------
        y: np.ndarray
            The output of the black box function.

        Raises
        ------
        NotImplementedError
            If the method is not implemented in the derived class.
        """
        raise NotImplementedError("abstract method")

    def terminate(self) -> None:
        """
        Terminate the black box optimization problem.
        """
        if hasattr(self, "inner_function"):
            self.inner_function.terminate()
        # if self.observer is not None:
        #     # NOTE: terminating a problem should gracefully end the observer process -> write the last state.
        #     self.observer.finish()
        return

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.terminate()

    def __del__(self):
        self.terminate()

    def __neg__(self):
        """
        Creates a new black box, where the objective function
        is the negative of the original one.
        """
        negative_black_box = NegativeBlackBox(self)
        return negative_black_box

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(L={self.info.max_sequence_length}, num_evaluations={self.num_evaluations})"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(L={self.info.max_sequence_length}, batch_size={self.batch_size}, parallelize={self.parallelize}, num_workers={self.num_workers}, evaluation_budget={self.evaluation_budget})>"


class NegativeBlackBox(AbstractBlackBox):
    """A wrapper for a black box that negates the objective function.

    If you construct a black-box function f for maximizing, then -f is
    a black-box function for minimizing. This class is a wrapper for
    implementing the latter.

    The only difference is that the __call__ method returns -f(x) instead
    of f(x). The _black_box method is the same as the original black box.
    """

    def __init__(self, f: AbstractBlackBox):
        self.f = f
        super().__init__(
            batch_size=f.batch_size,
            parallelize=f.parallelize,
            num_workers=f.num_workers,
            evaluation_budget=f.evaluation_budget,
        )

    def __call__(self, x, context=None):
        return -self.f.__call__(x, context)

    def _black_box(self, x, context=None):
        return self.f._black_box(x, context)

    def __str__(self) -> str:
        return f"NegativeBlackBox({self.f})"

    def __repr__(self) -> str:
        return f"<NegativeBlackBox({self.f})>"



"""This module contains utilities for loading PDB files and parsing them.
"""

from pathlib import Path
from typing import List

from Bio import PDB
from Bio.PDB.Residue import Residue
from Bio.SeqUtils import seq1


def parse_pdb_as_structure(
    path_to_pdb: Path, structure_name: str = "pdb", verbose: bool = False
) -> PDB.Structure.Structure:
    """Parses the PDB file at the given path and returns
    the structure.

    The parsing is done quietly by default,
    but you can set verbose=True to get some output.

    Parameters
    -----------
    path_to_pdb : Path
        The path to the PDB file.
    structure_name : str, optional
        The name of the structure (which is passed to the get_structure
        method of the PDBParser object). Defaults to "pdb".
    verbose : bool, optional
        If True, print the progress of the parsing. Defaults to False.

    Returns
    --------
    PDB.Structure.Structure
        The parsed structure.
    """
    parser = PDB.PDBParser(QUIET=not verbose)
    return parser.get_structure(structure_name, path_to_pdb)


def parse_pdb_as_residues(
    path_to_pdb: Path, structure_name: str = "pdb", verbose: bool = False
) -> List[Residue]:
    """
    Parse a PDB file and return a list of Residue objects.

    Parameters
    -----------
    path_to_pdb: Path
        The path to the PDB file.
    structure_name: str, optional
        The name of the structure. Defaults to "pdb".
    verbose: bool, optional
        Whether to print verbose output. Defaults to False.

    Returns
    --------
        residues: List[Residue]
            A list of Residue objects representing the parsed PDB file.
    """
    structure = parse_pdb_as_structure(path_to_pdb, structure_name, verbose)
    return list(structure.get_residues())


def parse_pdb_as_residue_strings(
    path_to_pdb: Path, structure_name: str = "pdb", verbose: bool = False
) -> List[str]:
    """
    Parse a PDB file and return a list of residue strings.

    Parameters
    ----------
    path_to_pdb : Path
        The path to the PDB file.
    structure_name : str, optional
        The name of the structure, by default "pdb".
    verbose : bool, optional
        Whether to print verbose output, by default False.

    Returns
    -------
    List[str]
        A list of residue strings.
    """
    residues = parse_pdb_as_residues(path_to_pdb, structure_name, verbose)

    return [seq1(res.get_resname()) for res in residues]



import numpy as np

from poli import objective_factory


def test_registering_white_noise():
    white_noise_problem = objective_factory.create(name="white_noise")
    f, x0 = white_noise_problem.black_box, white_noise_problem.x0
    _ = f(x0)
    f.terminate()


def test_registering_aloha():
    aloha_problem = objective_factory.create(name="aloha")
    f = aloha_problem.black_box
    x = np.array([list("ALOOF")])
    assert f(x) == 3
    f.terminate()



"""RFP objective factory and black box function."""

__author__ = "Simon Bartels"

from pathlib import Path

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.exceptions import FoldXNotFoundException
from poli.core.problem import Problem
from poli.core.util.isolation.instancing import instance_function_as_isolated_process
from poli.core.util.seeding import seed_python_numpy_and_torch
from poli.objective_repository.foldx_rfp_lambo import CORRECT_SEQ, PROBLEM_SEQ
from poli.objective_repository.foldx_rfp_lambo.information import AMINO_ACIDS


class FoldXRFPLamboBlackBox(AbstractBlackBox):
    def __init__(
        self,
        seed: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        batch_size: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        self.inverse_alphabet = {i + 1: AMINO_ACIDS[i] for i in range(len(AMINO_ACIDS))}
        self.inverse_alphabet[0] = "-"

        if not (Path.home() / "foldx" / "foldx").exists():
            raise FoldXNotFoundException(
                "FoldX wasn't found in ~/foldx/foldx. Please install it."
            )
        if not force_isolation:
            try:
                from poli.objective_repository.foldx_rfp_lambo.isolated_function import (
                    RFPWrapperIsolatedLogic,
                )

                self.inner_function = RFPWrapperIsolatedLogic(seed=seed)
            except (ImportError, FileNotFoundError):
                self.inner_function = instance_function_as_isolated_process(
                    name="foldx_rfp_lambo__isolated",
                )
        else:
            self.inner_function = instance_function_as_isolated_process(
                name="foldx_rfp_lambo__isolated",
            )

    def _black_box(self, x, context=None):
        return self.inner_function(x, context)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="foldx_rfp_lambo",
            max_sequence_length=244,
            aligned=False,
            fixed_length=False,
            deterministic=True,  # ?
            alphabet=AMINO_ACIDS,
            discrete=True,
            fidelity=None,
            padding_token="-",
        )


class FoldXRFPLamboProblemFactory(AbstractProblemFactory):
    def __init__(self):
        self.alphabet = AMINO_ACIDS
        self.problem_sequence = PROBLEM_SEQ
        self.correct_sequence = CORRECT_SEQ

    def create(
        self,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Creates a FoldX-RFP Lambo-specific black box instance, alongside initial
        observations.

        Parameters
        ----------
        seed : int, optional
            The seed value for random number generation, by default None.
        batch_size : int, optional
            The batch size for parallel evaluation, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize evaluation, by default False.
        num_workers : int, optional
            The number of workers for parallel evaluation, by default None.
        evaluation_budget : int, optional
            The evaluation budget, by default float("inf").
        force_isolation : bool, optional
            Should the problem
        Returns
        -------
        f : RaspBlackBox
            The FoldXRFP black box instance.
        x0 : np.ndarray
            The initial observations.
        y0 : np.ndarray
            The initial observations (i.e. the stability of the wildtypes).
        """
        # TODO: allow for bigger batch_sizes
        # For now (and because of the way the black box is implemented)
        # we only allow for batch_size=1
        if batch_size is None:
            batch_size = 1
        else:
            assert batch_size == 1

        # make problem reproducible
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        black_box = FoldXRFPLamboBlackBox(
            seed=seed,
            parallelize=parallelize,
            num_workers=num_workers,
            batch_size=batch_size,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )
        x0 = black_box.inner_function.x0

        return Problem(black_box, x0)



import pytest


@pytest.mark.poli__lambo
@pytest.mark.slow()
def test_foldx_rfp_lambo_runs():
    import numpy as np

    from poli import create

    # For now, we don't have automatic installation of lambo.
    # TODO: add automatic installation of lambo, and remove this
    # check.
    problem = create(name="foldx_rfp_lambo", seed=1)
    f = problem.black_box

    # Evaluating on the first base candidate
    first_base_candidate_and_mutation = np.array(
        [
            "AVIKEFMRFKVHMEG"
            "SMNGHEFEIEGEGEGR"
            "PYEGTQTAKLKVTKGG"
            "PLPFSWDILSPQFS"
            "RAFTKHPADIPDYYKQ"
            "SFPEGFKWERVMNFED"
            "GGAVTVTQDTSLED"
            "GTLIYKVKLRGTNFPP"
            "DGPVMQKKTMGWEAST"
            "ERLYPEDGVLKGDI"
            "KMALRLKDGGRYLADF"
            "KTTYKAKKPVQMPGAYN"
            "VDRKLDITSHNEDYTVV"
            "EQYERSEGRHSTG",
            "IVIKEFMRFKVHMEG"
            "SMNGHEFEIEGEGEGR"
            "PYEGTQTAKLKVTKGG"
            "PLPFSWDILSPQFS"
            "RAFTKHPADIPDYYKQ"
            "SFPEGFKWERVMNFED"
            "GGAVTVTQDTSLED"
            "GTLIYKVKLRGTNFPP"
            "DGPVMQKKTMGWEAST"
            "ERLYPEDGVLKGDI"
            "KMALRLKDGGRYLADF"
            "KTTYKAKKPVQMPGAYN"
            "VDRKLDITSHNEDYTVV"
            "EQYERSEGRHSTG",
        ]
    )
    assert np.isclose(
        f(first_base_candidate_and_mutation),
        np.array([[-10591.87684184, -61.8757], [-10634.23150497, -61.5511]]),
    ).all()



"""A wrapper around TDC's DRD3 docking objective function.

Prerequisites
-------------

There are several requirements for running this objective function. We expect you to

- have AutoDock Vina installed in the path,
- have the `prepare_receptor` binary from the ADFR suite installed and in the path,
- have the `poli__lambo` environment created.

The rest of this description shows you how to do all this:

Installing AutoDock Vina
========================

1. Download AutoDock Vina from the Center for Computational Structural Biology's website (https://vina.scripps.edu/downloads/). Uncompress them.

2. Add this to the path by including `export PATH=path/to/AutoDock_vina/bin:$PATH` in your `~/.bashrc` or `~/.zshrc`.

```bash
# In your ~/.bashrc or ~/.zshrc
export PATH=path/to/AutoDock_vina/bin:$PATH
```

Installing the ADFR suite
=========================

1. Download the installable files (https://ccsb.scripps.edu/adfr/downloads/). It's likely that you will have to run the `./install.sh` script inside the folder, and thus you might have to change its permissions for execution using `chmod +x`
2. After running `./install.sh`, you should be able to find `.../bin/prepare_receptor`.
3.  For the docking to run, `pyscreener` needs access to the `prepare_receptor` binary. However, adding all of the ADFR `bin` folder is sometimes problematic, since it has a version of Python inside. Thus, we recommend creating a symlink. Write this in your `~/.bashrc` or `~/.zshrc`:

```bash
# In your ~/.bashrc or ~/.zshrc
ln -sf /path/to/ADFR/bin/prepare_receptor /path/to/AutoDock_vina/bin
```

Create the `poli__lambo` environment
====================================

This can easily be done by running

.. code-block:: bash

        # From the base of the poli repo
        conda env create --file src/poli/objective_repository/ddr3_docking/environment.yml


We also need `lambo`'s tasks to be available in Python's path for `poli__lambo`:

.. code-block:: bash

            git clone https://github.com/samuelstanton/lambo    # For reference, we use 431b052
            cd lambo
            pip install -e .


In particular, we need
- `lambo.tasks.proxy_rfp.proxy_rfp.ProxyRFPTask`
- the rfp data: see `~/lambo/assets/fpbase`

And now you should be set! Running `register.py` and `query_example.py` should work.
"""



"""Registers the SASA FoldX black box and objective factory.

FoldX [1] is a simulator that allows for computing the difference
in free energy between a wildtype protein and a mutated protein.
We pair this with biopython [2] to compute the SASA of the mutated
protein.

[1] Schymkowitz, J., Borg, J., Stricher, F., Nys, R., Rousseau, F.,
    & Serrano, L. (2005). The FoldX web server: an online force field.
    Nucleic acids research, 33(suppl_2), W382-W388.
[2] Cock PA, Antao T, Chang JT, Chapman BA, Cox CJ, Dalke A, Friedberg I,
    Hamelryck T, Kauff F, Wilczynski B and de Hoon MJL (2009) Biopython:
    freely available Python tools for computational molecular biology and
    bioinformatics. Bioinformatics, 25, 1422-1423
"""

from pathlib import Path
from typing import List, Union

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.exceptions import FoldXNotFoundException
from poli.core.problem import Problem
from poli.core.util.isolation.instancing import get_inner_function
from poli.core.util.proteins.defaults import AMINO_ACIDS
from poli.core.util.seeding import seed_python_numpy_and_torch


class FoldXSASABlackBox(AbstractBlackBox):
    """
    A black box implementation for computing the solvent accessible surface area (SASA) score using FoldX.

    Parameters
    -----------
    wildtype_pdb_path : Union[Path, List[Path]]
        The path(s) to the wildtype PDB file(s). Default is None.
    experiment_id : str, optional
        The ID of the experiment. Default is None.
    tmp_folder : Path, optional
        The path to the temporary folder. Default is None.
    eager_repair : bool, optional
        Whether to perform eager repair. Default is False.
    verbose : bool, optional
        Flag indicating whether we print the output from FoldX. Default is False.
    batch_size : int, optional
        The batch size for parallel processing. Default is None.
    parallelize : bool, optional
        Whether to parallelize the computation. Default is False.
    num_workers : int, optional
        The number of workers for parallel processing. Default is None.
    evaluation_budget : int, optional
        The maximum number of function evaluations. Default is infinity.

    Notes
    -----
    We expect the user to have FoldX v5.0 installed and compiled.
    More specifically, we expect a binary file called foldx to be
    in the path ~/foldx/foldx.
    """

    def __init__(
        self,
        wildtype_pdb_path: Union[Path, List[Path]],
        experiment_id: str = None,
        tmp_folder: Path = None,
        eager_repair: bool = False,
        verbose: bool = False,
        batch_size: int = 1,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        self.wildtype_pdb_path = wildtype_pdb_path
        self.experiment_id = experiment_id
        self.tmp_folder = tmp_folder
        self.eager_repair = eager_repair
        self.verbose = verbose
        self.force_isolation = force_isolation

        if not (Path.home() / "foldx" / "foldx").exists():
            raise FoldXNotFoundException(
                "FoldX wasn't found in ~/foldx/foldx. Please install it."
            )

        inner_function = get_inner_function(
            isolated_function_name="foldx_sasa__isolated",
            class_name="FoldXSASAIsolatedLogic",
            module_to_import="poli.objective_repository.foldx_sasa.isolated_function",
            force_isolation=self.force_isolation,
            wildtype_pdb_path=self.wildtype_pdb_path,
            experiment_id=self.experiment_id,
            tmp_folder=self.tmp_folder,
            eager_repair=self.eager_repair,
            verbose=self.verbose,
        )
        self.wildtype_amino_acids = inner_function.wildtype_amino_acids

    def _black_box(self, x: np.ndarray, context: None) -> np.ndarray:
        """Computes the SASA score for a given mutation x.

        Runs the given input x and pdb files provided
        in the context through FoldX and returns the
        total SASA score.

        Parameters
        -----------
        x : np.ndarray
            The input array representing the mutations.
        context : None
            The context for the black box computation.

        Returns
        --------
        np.ndarray
            The computed SASA score(s) as a numpy array.
        """
        inner_function = get_inner_function(
            isolated_function_name="foldx_sasa__isolated",
            class_name="FoldXSASAIsolatedLogic",
            module_to_import="poli.objective_repository.foldx_sasa.isolated_function",
            force_isolation=self.force_isolation,
            quiet=True,
            wildtype_pdb_path=self.wildtype_pdb_path,
            experiment_id=self.experiment_id,
            tmp_folder=self.tmp_folder,
            eager_repair=self.eager_repair,
            verbose=self.verbose,
        )
        return inner_function(x, context)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="foldx_sasa",
            max_sequence_length=np.inf,
            aligned=False,
            fixed_length=False,
            deterministic=True,
            alphabet=AMINO_ACIDS,
            log_transform_recommended=False,
            discrete=True,
            fidelity=None,
            padding_token="",
        )


class FoldXSASAProblemFactory(AbstractProblemFactory):
    """
    Factory class for creating FoldX SASA (Solvent Accessible Surface Area) problems.

    Methods
    -------
    create:
        Creates a problem instance with the specified parameters.
    """

    def create(
        self,
        wildtype_pdb_path: Union[Path, List[Path]],
        experiment_id: str = None,
        tmp_folder: Path = None,
        eager_repair: bool = False,
        verbose: bool = False,
        seed: int = None,
        batch_size: int = 1,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Create a FoldXSASABlackBox object and compute the initial values of wildtypes.

        Parameters
        ----------
        wildtype_pdb_path : Union[Path, List[Path]]
            Path or list of paths to the wildtype PDB files.
        experiment_id : str, optional
            Identifier for the experiment.
        tmp_folder : Path, optional
            Path to the temporary folder for intermediate files.
        eager_repair : bool, optional
            Flag indicating whether to perform eager repair.
        verbose : bool, optional
            Flag indicating whether to print the output from FoldX.
            Default is False.
        seed : int, optional
            Seed for random number generators. If None is passed,
            the seeding doesn't take place.
        batch_size : int, optional
            Number of samples per batch for parallel computation.
        parallelize : bool, optional
            Flag indicating whether to parallelize the computation.
        num_workers : int, optional
            Number of worker processes for parallel computation.
        evaluation_budget : int, optional
            The maximum number of function evaluations. Default is infinity.

        Returns
        -------
        Tuple[AbstractBlackBox, np.ndarray, np.ndarray]
            A tuple containing the FoldXSASABlackBox object, the initial wildtype sequences, and the initial fitness values.

        Raises
        ------
        ValueError
            If wildtype_pdb_path is missing or has an invalid type.
        """
        # We start by seeding the RNGs
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        # We check whether the keyword arguments are valid
        if wildtype_pdb_path is None:
            raise ValueError(
                "Missing required argument wildtype_pdb_path. "
                "Did you forget to pass it to create()?"
            )

        if isinstance(wildtype_pdb_path, str):
            wildtype_pdb_path = [Path(wildtype_pdb_path.strip())]
        elif isinstance(wildtype_pdb_path, Path):
            wildtype_pdb_path = [wildtype_pdb_path]
        elif isinstance(wildtype_pdb_path, list):
            if isinstance(wildtype_pdb_path[0], str):
                wildtype_pdb_path = [Path(x.strip()) for x in wildtype_pdb_path]
            elif isinstance(wildtype_pdb_path[0], Path):
                pass
        else:
            raise ValueError(
                f"wildtype_pdb_path must be a string or a Path. Received {type(wildtype_pdb_path)}"
            )
        # By this point, we know that wildtype_pdb_path is a
        # list of Path objects.

        # We use the default alphabet if None was provided.
        # See ENCODING in foldx_utils.py

        f = FoldXSASABlackBox(
            wildtype_pdb_path=wildtype_pdb_path,
            experiment_id=experiment_id,
            tmp_folder=tmp_folder,
            eager_repair=eager_repair,
            verbose=verbose,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )

        # We need to compute the initial values of all wildtypes
        # in wildtype_pdb_path. For this, we need to specify x0,
        # a vector of wildtype sequences. These are padded to
        # match the maximum length with empty strings.
        wildtype_amino_acids_ = f.wildtype_amino_acids
        longest_wildtype_length = max([len(x) for x in wildtype_amino_acids_])

        wildtype_amino_acids = [
            amino_acids + [""] * (longest_wildtype_length - len(amino_acids))
            for amino_acids in wildtype_amino_acids_
        ]

        x0 = np.array(wildtype_amino_acids).reshape(
            len(wildtype_pdb_path), longest_wildtype_length
        )

        problem = Problem(f, x0)
        return problem



"""Utilities for loading up the cavity and downstream models for RaSP."""

from pathlib import Path
from typing import Tuple

import torch

from poli.core.util.proteins.rasp.inner_rasp.cavity_model import (
    CavityModel,
    DownstreamModel,
)
from poli.core.util.proteins.rasp.inner_rasp.helpers import init_lin_weights

THIS_DIR = Path(__file__).parent.resolve()
HOME_DIR = THIS_DIR.home()
RASP_DIR = HOME_DIR / ".poli_objectives" / "rasp"
RASP_DIR.mkdir(parents=True, exist_ok=True)


def load_cavity_and_downstream_models(
    device: str = "cpu",
) -> Tuple[CavityModel, DownstreamModel]:
    """Load the cavity and downstream models for RaSP.

    Parameters
    -----------
    device : str, optional
        The device to load the models on. Defaults to "cpu".

    Returns
    --------
    cavity_model_net : CavityModel
        The cavity model.
    ds_model_net : DownstreamModel
        The downstream model.
    """
    # DEVICE = "cpu"

    # TODO: Ask why this was implemented this way (i.e.
    # loading from a "best_cavity_model_path").
    # A transparent alternative would be to simply
    # load the model from the path itself.
    best_cavity_model_path = RASP_DIR / "cavity_model_15.pt"
    cavity_model_net = CavityModel(get_latent=True).to(device)
    cavity_model_net.load_state_dict(
        torch.load(f"{best_cavity_model_path}", map_location=device, weights_only=True)
    )
    cavity_model_net.eval()
    ds_model_net = DownstreamModel().to(device)
    ds_model_net.apply(init_lin_weights)
    ds_model_net.eval()

    return cavity_model_net, ds_model_net



from __future__ import annotations

from typing import List, Union

from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.problem import Problem


class AbstractBenchmark:
    problem_factories: List[AbstractProblemFactory]
    index: int = 0

    def __init__(
        self,
        seed: Union[int, None] = None,
        batch_size: Union[int, None] = None,
        parallelize: bool = False,
        num_workers: Union[int, None] = None,
        evaluation_budget: int = float("inf"),
    ) -> None:
        self.seed = seed
        self.batch_size = batch_size
        self.parallelize = parallelize
        self.num_workers = num_workers
        self.evaluation_budget = evaluation_budget

    def __len__(self) -> int:
        return len(self.problem_factories)

    def __getitem__(self, index: int) -> Problem:
        return self._initialize_problem(index)

    def __next__(self) -> Problem:
        if self.index < len(self.problem_factories):
            self.index += 1
            return self._initialize_problem(self.index - 1)
        else:
            raise StopIteration

    def _initialize_problem(self, index: int) -> Problem:
        # An abstract method that should be implemented by the subclasses.
        raise NotImplementedError

    @property
    def info(self) -> str:
        raise NotImplementedError

    @property
    def problem_names(self) -> List[str]:
        return [
            problem_factory.__module__.replace(
                "poli.objective_repository.", ""
            ).replace(".register", "")
            for problem_factory in self.problem_factories
        ]



"""
The isolation entry-point for Ehrlich, as implemented in Holo.
"""

from __future__ import annotations

import numpy as np
import torch
from holo.test_functions.closed_form._ehrlich import Ehrlich

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.registry import register_isolated_function
from poli.core.util.proteins.defaults import AMINO_ACIDS


class EhrlichIsolatedLogic(AbstractIsolatedFunction):
    """
    An isolated logic which uses Holo-bench's implementation
    of Ehrlich functions.
    """

    def __init__(
        self,
        sequence_length: int,
        motif_length: int,
        n_motifs: int,
        quantization: int | None = None,
        noise_std: float = 0.0,
        seed: int | None = None,
        epistasis_factor: float = 0.0,
        return_value_on_unfeasible: float = -np.inf,
        alphabet: list[str] = AMINO_ACIDS,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
    ):
        self.sequence_length = sequence_length
        self.motif_length = motif_length
        self.n_motifs = n_motifs
        self.epistasis_factor = epistasis_factor

        if seed is None:
            raise ValueError("The seed parameter must be set.")

        self.noise_std = noise_std
        self.quantization = quantization
        self.seed = seed
        self.return_value_on_unfeasible = return_value_on_unfeasible
        self.alphabet = alphabet
        self.parallelize = parallelize
        self.num_workers = num_workers
        self.evaluation_budget = evaluation_budget

        self.inner_ehrlich = Ehrlich(
            num_states=len(alphabet),
            dim=sequence_length,
            num_motifs=n_motifs,
            motif_length=motif_length,
            quantization=quantization,
            noise_std=noise_std,
            negate=False,  # We aim to maximize the function
            random_seed=seed,
        )

    def __call__(self, x: np.ndarray, context: None) -> np.ndarray:
        # First, we transform the strings into integers using the alphabet
        batch_size = x.shape[0]
        if len(x.shape) > 1:
            # Flattening each element
            x = np.array(["".join(x_i) for x_i in x])
        x_ = np.array([[self.alphabet.index(c) for c in s] for s in x.flatten()])

        values = self.inner_ehrlich(torch.from_numpy(x_)).numpy(force=True)
        values[values == -np.inf] = self.return_value_on_unfeasible

        return values.reshape(batch_size, 1)

    def initial_solution(self, n_samples: int = 1) -> np.ndarray:
        return self.inner_ehrlich.initial_solution(n=n_samples).numpy(force=True)

    @property
    def optimal_solution(self):
        return self.inner_ehrlich.optimal_solution().numpy(force=True)

    @property
    def random_solution(self):
        return self.inner_ehrlich.random_solution().numpy(force=True)

    @property
    def transition_matrix(self):
        return self.inner_ehrlich.transition_matrix.numpy(force=True)


if __name__ == "__main__":
    register_isolated_function(
        EhrlichIsolatedLogic,
        name="ehrlich_holo__isolated",
        conda_environment_name="poli__ehrlich_holo",
    )



"""
This module tests the utilities related to proteins,
which can be found in poli.core.util.proteins.

Since this code is only intended to run inside the
poli__protein environment, we add a skip to the
whole module if the import fails.
"""

from pathlib import Path

import pytest

try:
    from poli.core.util.proteins.pdb_parsing import (
        parse_pdb_as_residue_strings,
        parse_pdb_as_residues,
    )
except ImportError:
    pytest.skip(
        "Could not import the protein utilities for parsing. ", allow_module_level=True
    )

try:
    from poli.core.util.proteins.mutations import (
        find_closest_wildtype_pdb_file_to_mutant,
        mutations_from_wildtype_residues_and_mutant,
    )
except ImportError:
    pytest.skip(
        "Could not import the protein utilities for mutations. ",
        allow_module_level=True,
    )

THIS_DIR = Path(__file__).parent.resolve()


@pytest.mark.poli__protein
class TestClosestPDBFilesToMutation:
    wildtype_pdb_paths = [
        THIS_DIR / "3ned.pdb",
        THIS_DIR / "2vae.pdb",
    ]

    wildtype_strings = [
        parse_pdb_as_residue_strings(pdb_path) for pdb_path in wildtype_pdb_paths
    ]

    wildtype_residues = [
        parse_pdb_as_residues(pdb_path) for pdb_path in wildtype_pdb_paths
    ]

    one_mutation_to_3ned = "".join(wildtype_strings[0])
    one_mutation_to_3ned = "A" + one_mutation_to_3ned[1:]

    def test_finding_closest_pdb_to_mutation(self):
        """
        Loads up two PDB files, computes a couple of mutations
        of one, and then finds the closest PDB file to the
        mutated one.
        """
        # We check whether the closest PDB file to the mutated
        # one is 3ned.
        closest_pdb_path = find_closest_wildtype_pdb_file_to_mutant(
            wildtype_pdb_files=self.wildtype_pdb_paths,
            mutated_residue_string=self.one_mutation_to_3ned,
        )
        assert closest_pdb_path == self.wildtype_pdb_paths[0]

    def test_finding_closest_pdb_is_agnostic_to_lowercase(self):
        """
        Loads up two PDB files, computes a couple of mutations
        of one, and then finds the closest PDB file to the
        mutated one.
        """
        # We check whether the closest PDB file to the mutated
        # one is 3ned, but now we use lowercase.
        closest_pdb_path = find_closest_wildtype_pdb_file_to_mutant(
            wildtype_pdb_files=self.wildtype_pdb_paths,
            mutated_residue_string=self.one_mutation_to_3ned.lower(),
        )
        assert closest_pdb_path == self.wildtype_pdb_paths[0]

    def test_computing_mutations_from_wildtype_and_mutant(self):
        """
        We test whether we can compute the mutations from
        a wildtype and a mutant.
        """
        mutations = mutations_from_wildtype_residues_and_mutant(
            wildtype_residues=self.wildtype_residues[0],
            mutated_residue_string=self.one_mutation_to_3ned,
        )

        assert mutations == ["EA1A"]

    def test_computing_mutations_is_agnostic_to_lowercase(self):
        """
        We test whether we can compute the mutations from
        a wildtype and a mutant.
        """
        mutations = mutations_from_wildtype_residues_and_mutant(
            wildtype_residues=self.wildtype_residues[0],
            mutated_residue_string=self.one_mutation_to_3ned.lower(),
        )

        assert mutations == ["EA1A"]



"""External observer, which can be run in an isolated process."""

from typing import Any

import numpy as np

from poli.core.black_box_information import BlackBoxInformation
from poli.core.registry import _DEFAULT, _OBSERVER, config
from poli.core.util.abstract_observer import AbstractObserver
from poli.core.util.inter_process_communication.process_wrapper import ProcessWrapper


class ExternalObserver(AbstractObserver):
    """An external version of the observer class to be instantiated in isolated processes.

    Parameters
    ----------
    observer_name : str, optional
        The name of the observer. If not provided, the default observer name will be used.
    **kwargs_for_observer
        Additional keyword arguments to be passed to the observer.

    Methods
    -------
    observe(x, y, context=None)
        Sends the observation to the observer process and verifies if it was logged correctly.
    initialize_observer(setup_info, caller_info, x0, y0, seed)
        Starts the observer process and sends the setup information.
    finish()
        Closes the observer process.
    __getattr__(__name)
        Retrieves the attribute of the underlying observer.
    """

    def __init__(self, observer_name: str = None, **kwargs_for_observer):
        """
        Initialize the ExternalObserver object.

        Parameters
        ----------
        observer_name : str, optional
            The name of the observer. If not provided, the default observer will be used.
        **kwargs_for_observer
            Additional keyword arguments to be passed to the observer.
        """
        if observer_name is None:
            observer_name = _DEFAULT

        self.observer_name = observer_name

        self.observer_script = config[_OBSERVER][observer_name]
        self.process_wrapper = None
        self.kwargs_for_observer = kwargs_for_observer

    def observe(self, x: np.ndarray, y: np.ndarray, context=None) -> None:
        """
        Observe the given data points.

        Parameters
        ----------
        x: np.ndarray
            The input data points.
        y: np.ndarray
            The output data points.
        context: object
            Additional context for the observation.

        Raises
        -------
        Exception:
            If the underlying observer process raises an exception.
        """

        # We send the observation
        self.process_wrapper.send(["OBSERVATION", x, y, context])

        # And we make sure the process received and logged it correctly
        msg_type, *msg = self.process_wrapper.recv()
        if msg_type == "EXCEPTION":
            e, tb = msg
            print(tb)
            raise e

        # else, it was a successful observation

    def initialize_observer(
        self, setup_info: BlackBoxInformation, caller_info, seed
    ) -> str:
        """
        Initialize the observer.

        Parameters
        ----------
        problem_setup_info : ProblemSetupInformation
            The information about the problem setup.
        caller_info : object
            The information about the caller.
        seed : int
            The seed value for random number generation.

        Returns
        -------
        observer_info: object
            Relevant information returned by the underlying observer's
            initialize_observer method.

        Raises
        ------
        Exception:
            Any exception raised by the underlying observer.
        ValueError:
            If the message type received from the observer process is unknown.
        """
        # start observer process
        # VERY IMPORTANT: the observer script MUST accept port and password as arguments
        self.process_wrapper = ProcessWrapper(self.observer_script)

        # send setup information
        self.process_wrapper.send(
            [setup_info, caller_info, seed, self.kwargs_for_observer]
        )

        # wait for logger handle
        msg_type, *msg = self.process_wrapper.recv()
        if msg_type == "SETUP":
            observer_info = msg[0]
        elif msg_type == "EXCEPTION":
            e, tb = msg
            print(tb)
            raise e
        else:
            raise ValueError("Unknown message type from observer process: " + msg_type)

        # forward to objective factory
        return observer_info

    def log(self, algorithm_info: dict):
        # We send the observation
        self.process_wrapper.send(["LOG", algorithm_info])

        # And we make sure the process received and logged it correctly
        msg_type, *msg = self.process_wrapper.recv()
        if msg_type == "EXCEPTION":
            e, tb = msg
            print(tb)
            raise e

        # else, it was a successful observation

    def finish(self) -> None:
        """Finish the external observer process.

        This method sends a "QUIT" message to the process wrapper and closes it.
        """
        if self.process_wrapper is not None:
            self.process_wrapper.send(["QUIT", None])
            self.process_wrapper.close()
            self.process_wrapper = None

    def __getattr__(self, __name: str) -> Any:
        """Get an attribute of the underlying observer if it exists.

        Asks for the attribute of the underlying
        black-box function by sending a message
        to the process w. the msg_type "ATTRIBUTE".
        """
        self.process_wrapper.send(["ATTRIBUTE", __name])
        msg_type, *msg = self.process_wrapper.recv()
        if msg_type == "EXCEPTION":
            e, tb = msg
            print(tb)
            raise e
        else:
            assert msg_type == "ATTRIBUTE"
            attribute = msg[0]
            return attribute

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(name={self.observer_name})"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(name={self.observer_name}, script_location={self.observer_script})>"



"""
Stanton et al.'s [1] implementation of Ehrlich functions using Holo-bench.

References
----------
[1] Stanton, S., Alberstein, R., Frey, N., Watkins, A., & Cho, K. (2024).
    Closed-Form Test Functions for Biophysical Sequence Optimization Algorithms.
    arXiv preprint arXiv:2407.00236. https://arxiv.org/abs/2407.00236
"""

from __future__ import annotations

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.isolation.instancing import get_inner_function
from poli.core.util.proteins.defaults import AMINO_ACIDS
from poli.core.util.seeding import seed_python_numpy_and_torch


class EhrlichHoloBlackBox(AbstractBlackBox):
    """
    Ehrlich functions were proposed by Stanton et al. [1] as a quick-and-easy
    alternative for testing discrete sequence optimizers (with protein
    optimization in mind). They are deviced to

    (i) be easy to query,
    (ii) have feasible and unfeasible sequences,
    (iii) have uninformative random samples (i.e. randomly sampling and evaluating should not be competitive, as many of these should be unfeasible).
    (iv) be maximized when certain motifs are present in the sequence. These motifs can be long-range within the sequence, and are meant to be non-additive.

    Check the references for details on the implementation.

    Parameters
    ----------
    sequence_length : int
        The length of the sequence to be optimized. This length is fixed, and
        _only_ sequences of this length are considered.
    motif_length : int
        The length of the motifs.
    n_motifs : int
        The number of motifs.
    quantization : int, optional
        The quantization parameter. This parameter must be between 1 and the
        motif length, and the motif length must be divisible by the quantization.
        By default, it is None (which corresponds to the motif length).
    noise_std : float, optional
        The noise that gets injected into botorch's SyntheticTestFunction.
        By default, it is 0.0.
    seed : int, optional
        The seed for the random number generator. By default, it is None
        (i.e. a random seed is set using np.random.randint(0, 1000)).
    epistasis_factor : float, optional
        The epistasis factor. By default, it is 0.0.
    return_value_on_unfeasible : float, optional
        The value to be returned when an unfeasible sequence is evaluated.
        By default, it is -np.inf.
    alphabet : list of str, optional
        The alphabet to be used for the sequences. By default, it is the
        of 20 amino acids.
    batch_size : int, optional
        The batch size for the black box. By default, it is None (i.e. all
        sequences are evaluated in a vectorized way).
    parallelize : bool, optional
        Whether to parallelize the evaluation of the black box. By default,
        it is False.
    num_workers : int, optional
        The number of processors used in parallelization.
    evaluation_budget : int, optional
        The evaluation budget for the black box. By default, it is infinite.

    References
    ----------
    [1] Stanton, S., Alberstein, R., Frey, N., Watkins, A., & Cho, K. (2024).
        Closed-Form Test Functions for Biophysical Sequence Optimization Algorithms.
        arXiv preprint arXiv:2407.00236. https://arxiv.org/abs/2407.00236

    """

    def __init__(
        self,
        sequence_length: int,
        motif_length: int,
        n_motifs: int,
        quantization: int | None = None,
        noise_std: float = 0.0,
        seed: int = None,
        epistasis_factor: float = 0.0,
        return_value_on_unfeasible: float = -np.inf,
        alphabet: list[str] = AMINO_ACIDS,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ):
        super().__init__(batch_size, parallelize, num_workers, evaluation_budget)
        self.alphabet = alphabet
        self.sequence_length = sequence_length
        self.return_value_on_unfeasible = return_value_on_unfeasible

        if seed is None:
            # In the case of Ehrlich, it's important we
            # set the seed here, as it will be used in the inner function.
            seed = np.random.randint(1, 1000)

        seed_python_numpy_and_torch(seed)
        self.seed = seed

        if motif_length * n_motifs > sequence_length:
            raise ValueError(
                "The total length of the motifs is greater than the sequence length."
            )

        if quantization is None:
            quantization = motif_length

        if not (1 <= quantization <= motif_length) or motif_length % quantization != 0:
            raise ValueError(
                "The quantization parameter must be between 1 and the motif length, "
                "and the motif length must be divisible by the quantization."
            )

        self.motif_length = motif_length
        self.n_motifs = n_motifs
        self.quantization = quantization

        self.inner_function = get_inner_function(
            isolated_function_name="ehrlich_holo__isolated",
            class_name="EhrlichIsolatedLogic",
            module_to_import="poli.objective_repository.ehrlich_holo.isolated_function",
            force_isolation=force_isolation,
            sequence_length=sequence_length,
            motif_length=motif_length,
            n_motifs=n_motifs,
            quantization=quantization,
            noise_std=noise_std,
            seed=self.seed,
            epistasis_factor=epistasis_factor,
            return_value_on_unfeasible=return_value_on_unfeasible,
            alphabet=alphabet,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

    def initial_solution(self, n_samples: int = 1) -> np.ndarray:
        # This is a sequence of ints.
        initial_solution_as_ints = self.inner_function.initial_solution(n_samples)

        # We convert it to a sequence of strings.
        if n_samples == 1:
            return np.array(
                ["".join([self.alphabet[i] for i in initial_solution_as_ints])]
            )
        else:
            return np.array(
                [
                    "".join([self.alphabet[i] for i in z_i])
                    for z_i in initial_solution_as_ints
                ]
            )

    def random_solution(self) -> np.ndarray:
        random_solution_as_ints = self.inner_function.random_solution

        return np.array(["".join([self.alphabet[i] for i in random_solution_as_ints])])

    def optimal_solution(self) -> np.ndarray:
        optimal_solution_as_ints = self.inner_function.optimal_solution

        return np.array(["".join([self.alphabet[i] for i in optimal_solution_as_ints])])

    def transition_matrix(self) -> np.ndarray:
        return self.inner_function.transition_matrix

    def _black_box(self, x: np.ndarray, context=None) -> np.ndarray:
        """
        Evaluates the sequences in x by checking maximal matches and multiplying.
        """
        return self.inner_function(x, context=context)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="ehrlich_holo",
            max_sequence_length=self.sequence_length,
            aligned=True,
            fixed_length=True,
            deterministic=True,
            alphabet=self.alphabet,
            log_transform_recommended=False,
            discrete=True,
            padding_token="",
        )


class EhrlichHoloProblemFactory(AbstractProblemFactory):
    """
    A factory for creating Ehrlich functions.

    References
    ----------
    [1] Stanton, S., Alberstein, R., Frey, N., Watkins, A., & Cho, K. (2024).
        Closed-Form Test Functions for Biophysical Sequence Optimization Algorithms.
        arXiv preprint arXiv:2407.00236. https://arxiv.org/abs/2407.00236
    """

    def create(
        self,
        sequence_length: int,
        motif_length: int,
        n_motifs: int,
        quantization: int | None = None,
        noise_std: float = 0.0,
        seed: int = None,
        epistasis_factor: float = 0.0,
        return_value_on_unfeasible: float = -np.inf,
        alphabet: list[str] = AMINO_ACIDS,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Creates an Ehrlich function problem (containing an Ehrlich black box and
        an initial condition).

        Parameters
        ----------
        sequence_length : int
            The length of the sequence to be optimized. This length is fixed, and
            _only_ sequences of this length are considered.
        motif_length : int
            The length of the motifs.
        n_motifs : int
            The number of motifs.
        quantization : int, optional
            The quantization parameter. This parameter must be between 1 and the
            motif length, and the motif length must be divisible by the quantization.
            By default, it is None (which corresponds to the motif length).
        noise_std : float, optional
            The noise that gets injected into botorch's SyntheticTestFunction.
            By default, it is 0.0.
        seed : int, optional
            The seed for the random number generator. By default, it is None
            (i.e. a random seed is set using np.random.randint(0, 1000)).
        epistasis_factor : float, optional
            The epistasis factor. By default, it is 0.0.
        return_value_on_unfeasible : float, optional
            The value to be returned when an unfeasible sequence is evaluated.
            By default, it is -np.inf.
        alphabet : list of str, optional
            The alphabet to be used for the sequences. By default, it is the
            of 20 amino acids.
        batch_size : int, optional
            The batch size for the black box. By default, it is None (i.e. all
            sequences are evaluated in a vectorized way).
        parallelize : bool, optional
            Whether to parallelize the evaluation of the black box. By default,
            it is False.
        num_workers : int, optional
            The number of processors used in parallelization.
        evaluation_budget : int, optional
            The evaluation budget for the black box. By default, it is infinite.

        References
        ----------
        [1] Stanton, S., Alberstein, R., Frey, N., Watkins, A., & Cho, K. (2024).
            Closed-Form Test Functions for Biophysical Sequence Optimization Algorithms.
            arXiv preprint arXiv:2407.00236. https://arxiv.org/abs/2407.00236
        """
        if seed is not None:
            seed_python_numpy_and_torch(seed)

        f = EhrlichHoloBlackBox(
            sequence_length=sequence_length,
            motif_length=motif_length,
            n_motifs=n_motifs,
            quantization=quantization,
            noise_std=noise_std,
            seed=seed,
            epistasis_factor=epistasis_factor,
            return_value_on_unfeasible=return_value_on_unfeasible,
            alphabet=alphabet,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )
        x0 = f.initial_solution()

        return Problem(f, x0)



from pathlib import Path

import numpy as np

from poli.core.data_package import DataPackage


class RFPRaspSupervisedDataPackage(DataPackage):
    def __init__(self):
        PROTEIN_DATA_PACKAGES_DIR = Path(__file__).parent
        sequences = np.loadtxt(
            PROTEIN_DATA_PACKAGES_DIR / "rfp_sequences.txt", dtype=str
        )
        rasp_scores = np.loadtxt(PROTEIN_DATA_PACKAGES_DIR / "rfp_rasp_scores.txt")
        padding_token = ""
        max_sequence_length = max(len(sequence) for sequence in sequences)
        unsupervised_data = np.array(
            [
                list(sequence) + [padding_token] * (max_sequence_length - len(sequence))
                for sequence in sequences
            ]
        )
        supervised_data = unsupervised_data, rasp_scores.reshape(-1, 1)

        super().__init__(unsupervised_data, supervised_data)



"""FoldX interface for measuring stability and SASA.

This module implements a `FoldxInterface` for querying
foldx's repairing and simulating models [1]. This implementation
is heavily inspired by the way LaMBO uses FoldX [2].

If this module is imported from a script, it will
automatically check that the foldx files are in the
expected location, and raise an error if they are not.

References
----------
[1] The FoldX web server: an online force field.
    Schymkowitz, J., Borg, J., Stricher, F., Nys, R.,
    Rousseau, F., & Serrano, L. (2005).  Nucleic acids research,
    33(suppl_2), W382-W388.

[2] “Accelerating Bayesian Optimization for Biological Sequence
    Design withDenoising Autoencoders.” Stanton, Samuel, Wesley Maddox,
    Nate Gruver, Phillip Maffettone, Emily Delaney, Peyton Greenside,
    and Andrew Gordon Wilson.  arXiv, July 12, 2022.
    http://arxiv.org/abs/2203.12742.
    https://github.com/samuelstanton/lambo

"""

import logging
import os
import shutil
import subprocess
from pathlib import Path
from typing import List, Union

from Bio.PDB import SASA
from Bio.PDB.Residue import Residue
from Bio.SeqUtils import seq1
from pdbtools.pdb_delhetatm import run as pdb_delhetatm_run

from poli.core.util.proteins.mutations import (
    mutations_from_wildtype_residues_and_mutant,
)
from poli.core.util.proteins.pdb_parsing import (
    parse_pdb_as_residues,
    parse_pdb_as_structure,
)

# Making sure that the foldx executable is where
# we expect
PATH_TO_FOLDX_FILES = Path().home() / "foldx"
if not PATH_TO_FOLDX_FILES.exists():
    raise FileNotFoundError(
        "Please download FoldX and place it in your home directory. \n"
        "We expect it to find the following files: \n"
        "   - the binary at: ~/foldx/foldx  \n"
    )

if not (PATH_TO_FOLDX_FILES / "foldx").exists():
    raise FileNotFoundError(
        "Please compile FoldX and place it in your home directory as 'foldx'. \n"
        "We expect it to find the following files: \n"
        "   - the binary at: ~/foldx/foldx  \n"
    )


class FoldxInterface:
    """
    A class for interacting with FoldX, a protein engineering software.

    Parameters
    ----------
    working_dir : Union[Path, str]
        The working directory where FoldX files and output will be stored.

    Methods
    -------
    repair(pdb_file, remove_and_rename=False, pH=7.0, remove_heteroatoms=True)
        Repairs a PDB file with FoldX, overwriting the original file if remove_and_rename is True.
    _repair_if_necessary_and_provide_path(pdb_file)
        Repairs a PDB file if necessary and returns the path of the repaired PDB.
    _simulate_mutations(pdb_file, mutations=None)
        Simulates mutations on a PDB file with FoldX.
    _read_energy(pdb_file)
        Reads the energy from a FoldX results file.
    _compute_sasa(pdb_file)
        Computes the SASA (solvent-accessible surface area) from a FoldX results file.
    compute_stability(pdb_file, mutations=None)
        Computes the stability of a protein structure using FoldX.
    compute_sasa(pdb_file, mutations=None)
        Computes the SASA (solvent-accessible surface area) of a protein structure using FoldX.
    compute_stability_and_sasa(pdb_file, mutations=None)
        Computes the stability and SASA (solvent-accessible surface area) of a protein structure using FoldX in a single run.
    copy_foldx_files(pdb_file)
        Copies the necessary FoldX files to the working directory.
    write_mutations_to_file(wildtype_resiudes, mutations, output_dir)
        Writes the list of mutations to a file in the given directory.

    Attributes
    ----------
    working_dir : Union[Path, str]
        The working directory for FoldX.
    verbose : bool
        If True, the FoldX output will be printed to stdout.

    Notes
    -----
    This class expects you to use the binary for FoldX v.5.
    Previous versions relied on a "rotabase.txt" file, which
    is no longer used.
    """

    def __init__(self, working_dir: Union[Path, str], verbose: bool = False):
        """
        Initialize the FoldX object.

        Parameters
        ----------
        working_dir : Union[Path, str]
            The working directory for FoldX.
        verbose : bool, optional
            If True, the FoldX output will be printed to stdout. Default is False.
        """
        if isinstance(working_dir, str):
            working_dir = Path(working_dir)

        self.working_dir = working_dir
        self.verbose = verbose

        if not verbose:
            self.output = subprocess.DEVNULL
        else:
            self.output = None

    def repair(
        self,
        pdb_file: Union[str, Path],
        remove_and_rename: bool = False,
        pH: float = 7.0,
        remove_heteroatoms: bool = True,
    ) -> None:
        """
        Repairs a PDB file with FoldX, overwriting
        the original file if remove_and_rename is True (default: False).

        Parameters
        ----------
        pdb_file : Union[str, Path]
            The path to the PDB file to be repaired.
        remove_and_rename : bool, optional
            If True, the original file will be removed and the repaired file will be renamed to the original file name.
            Default is False.
        pH : float, optional
            The pH value for the repair process. Default is 7.0.
        remove_heteroatoms : bool, optional
            If True, heteroatoms will be removed from the repaired PDB file using pdbtools.
            Default is True.

        Raises
        ------
        RuntimeError
            If FoldX fails to repair the PDB file.

        Notes:
        ------
        This method repairs a PDB file using FoldX. It overwrites the original file if remove_and_rename is True.
        """
        # Make sure the pdb file is a path
        if isinstance(pdb_file, str):
            pdb_file = Path(pdb_file)

        # Make sure the relevant files are in the
        # working directory
        self.copy_foldx_files(pdb_file)

        # The command to run
        command = [
            str(PATH_TO_FOLDX_FILES / "foldx"),
            "--command=RepairPDB",
            "--pdb",
            f"{pdb_file.stem}.pdb",
            "--water",
            "-CRYSTAL",
            "--pH",
            f"{pH}",
        ]

        # Running it in the working directory
        try:
            subprocess.run(
                command,
                cwd=self.working_dir,
                check=True,
                stdout=self.output,
                stderr=self.output,
            )
        except subprocess.CalledProcessError as e:
            raise RuntimeError(
                f"FoldX failed to repair the pdb file {pdb_file}. " + "\n"
                "This can happen because: \n"
                "- the foldx license (which has a 1-year turnaround) expired, or"
                "- the PDB file is not repaired. \n"
                "If you want to dig deeper, you can check the FoldX output in the working directory: "
                f"{self.working_dir}."
            ) from e

        # Checking that the file was generated
        repaired_pdb_file = self.working_dir / f"{pdb_file.stem}_Repair.pdb"
        assert repaired_pdb_file.exists(), (
            f"FoldX did not generate the expected repaired pdb file {repaired_pdb_file}. "
            f"Please check the working directory: {self.working_dir}. "
        )

        # If remove heteroatoms is True, we remove them
        # using pdbtools
        if remove_heteroatoms:
            # We load up the repaired file
            with open(repaired_pdb_file) as f:
                lines = f.readlines()

            deleting_heteroatoms_result = pdb_delhetatm_run(lines)

            # We write the result to the same file
            with open(repaired_pdb_file, "w") as f:
                f.writelines(deleting_heteroatoms_result)

        # Removing the old pdb file, and renaming the repaired one
        if remove_and_rename:
            shutil.rmtree(self.working_dir / f"{pdb_file.stem}.pdb")
            shutil.move(
                self.working_dir / f"{pdb_file.stem}_Repair.pdb",
                self.working_dir / f"{pdb_file.stem}.pdb",
            )

    def _repair_if_necessary_and_provide_path(self, pdb_file: Path) -> Path:
        """
        If the pdb_file's name doesn't end in "_Repair.pdb",
        then we repair it and return the path of the repaired
        pdb. Otherwise, we return the same path as the input.

        Parameters
        ----------
        pdb_file : Path
            The path to the PDB file to be repaired.

        Returns
        -------
        repaired_path: Path
            The path to the repaired PDB file.

        """
        # Make sure that we don't have a repaired pdb file
        # in the working directory (which is usually a cache)
        if (self.working_dir / f"{pdb_file.stem}_Repair.pdb").exists():
            logging.warning(
                f"Found a repaired pdb file in the cache for {pdb_file.stem}. Using it instead of repairing."
            )
            return self.working_dir / f"{pdb_file.stem}_Repair.pdb"

        # If the file's already fixed, then we don't need to
        # do anything. Else, we repair it.
        if "_Repair" in pdb_file.name:
            return pdb_file
        else:
            self.repair(pdb_file)
            return self.working_dir / f"{pdb_file.stem}_Repair.pdb"

    def _simulate_mutations(self, pdb_file: Path, mutations: List[str] = None) -> None:
        """Simulates mutations, starting from a wildtype PDB file.

        This method simulates mutations on a PDB file with FoldX.
        The list of mutations must be as expected by the
        individual_list.txt, i.e. a list of strings of the following
        form:
            - the first letter is the original residue,
            - the second letter is the chain ID,
            - the third letter is the position of the mutation,
            - the fourth letter is the mutant residue.
        e.g. ["MA0A"] means that the first residue in the chain
        is mutated from M to A.

        Parameters
        ----------
        pdb_file : Path
            The path to the PDB file to be repaired.
        mutations : List[str], optional
            The list of mutations to simulate. If None, we simulate
            the wildtype. Default is None.

        Raises
        ------
        AssertionError
            If the number of mutations is not 0 or 1.
        RuntimeError
            If FoldX fails to simulate the mutations.

        Notes:
        ------
        This method simulates mutations on a PDB file with FoldX.
        The list of mutations must be as expected by the
        individual_list.txt, i.e. a list of strings of the following
        form:
            - the first letter is the original residue,
            - the second letter is the chain ID,
            - the third letter is the position of the mutation,
            - the fourth letter is the mutant residue.
        e.g. ["MA0A"] means that the first residue in the chain
        is mutated from M to A.
        """
        # TODO: add support for multiple mutations.
        if mutations is not None:
            assert len(mutations) in [
                0,
                1,
            ], "We only support single mutations for now. "

        # We load up the residues of the wildtype
        wildtype_residues = parse_pdb_as_residues(pdb_file)

        # If there are no mutations, we still need to pass
        # a dummy mutation to foldx, so we "mutate" the first
        # residue in the wildtype string. This can be achieved
        # by passing the wildtype string as the mutated string.
        if mutations is None or len(mutations) == 0:
            wildtype_string = [seq1(res.get_resname()) for res in wildtype_residues]
            mutations = ["".join(wildtype_string)]

        # We write the individual_list.txt in the working directory
        self.write_mutations_to_file(wildtype_residues, mutations, self.working_dir)

        # We copy the relevant files into the working directory
        self.copy_foldx_files(pdb_file)

        # We run foldx
        # This generates the Raw_*.fxout file, which
        # contains the stability. If that file wasn't
        # generated, we raise an error.
        foldx_command = [
            str(PATH_TO_FOLDX_FILES / "foldx"),
            "--command=BuildModel",
            "--pdb",
            f"{pdb_file.stem}.pdb",
            "--mutant-file",
            "individual_list.txt",
            "--water",
            "-CRYSTAL",
            "--pH",
            "7.0",
        ]
        try:
            subprocess.run(
                foldx_command,
                cwd=self.working_dir,
                check=True,
                stdout=self.output,
                stderr=self.output,
            )
        except subprocess.CalledProcessError as e:
            raise RuntimeError(
                f"FoldX failed to simulate the mutations on the pdb file {pdb_file}. "
                f"Please check the working directory: {self.working_dir}. "
            ) from e

        results_dir = self.working_dir / f"Raw_{pdb_file.stem}.fxout"
        mutated_structure_dir = self.working_dir / f"{pdb_file.stem}_1.pdb"
        assert results_dir.exists(), (
            f"FoldX did not generate the expected results file {results_dir}. "
            f"Please check the working directory: {self.working_dir}. "
        )
        assert mutated_structure_dir.exists(), (
            f"FoldX did not generate the expected mutated pdb file {mutated_structure_dir}. "
            f"Please check the working directory: {self.working_dir}. "
        )

    def _read_energy(self, pdb_file: Path) -> float:
        """
        This method reads the energy from a FoldX results file,
        assuming that there was a single mutation.
        TODO: add support for multiple mutations.

        Parameters
        ----------
        pdb_file : Path
            The path to the PDB file to be repaired.

        Returns
        -------
        energy: float
            The change of energy (ddG) of the mutated structure.

        Raises
        ------
        AssertionError
            If the results file was not generated.
        """
        assert (
            self.working_dir / f"Raw_{pdb_file.stem}.fxout"
        ).exists(), f"FoldX did not generate the expected results file {self.working_dir / f'Raw_{pdb_file.stem}.fxout'}. "
        with open(self.working_dir / f"Raw_{pdb_file.stem}.fxout") as f:
            lines = f.readlines()

        energy = float(lines[-2].split()[1])
        return energy

    def _compute_sasa(self, pdb_file: Path) -> float:
        """
        This method computes the SASA from a FoldX results file,
        assuming that there was a single mutation.

        Parameters
        ----------
        pdb_file : Path
            The path to the PDB file to be repaired.

        Returns
        -------
        sasa_score: float
            The SASA score of the mutated structure.
        """
        mutated_structure = parse_pdb_as_structure(
            self.working_dir / f"{pdb_file.stem}_1.pdb", structure_name="pdb_mutated"
        )

        sasa_computer = SASA.ShrakeRupley()

        # This computes the sasa score, and attaches
        # it to the structure.
        sasa_computer.compute(mutated_structure, level="S")

        return mutated_structure.sasa

    def compute_stability(self, pdb_file: Path, mutations: List[str] = None) -> float:
        """
        Compute the stability of a protein structure using FoldX.

        Parameters
        ----------
        pdb_file : Path
            The path to the PDB file of the protein structure.
        mutations : List[str], optional
            A list of mutations to be simulated. Only single mutations are supported. Pass no mutations to compute the energy of the wildtype.

        Returns
        -------
        float
            The stability of the protein structure (defined as the negative
            change in energy).

        Raises
        ------
        AssertionError
            If the number of mutations is not 0 or 1.

        """
        # if not (self.working_dir / f"Raw_{pdb_file.stem}.fxout").exists():
        if mutations is not None:
            assert len(mutations) in [
                0,
                1,
            ], "We only support single mutations for now. Pass no mutations if you want to compute the energy of the wildtype."

        self._simulate_mutations(pdb_file, mutations)

        stability = -self._read_energy(pdb_file)
        return stability

    def compute_sasa(self, pdb_file: Path, mutations: List[str] = None) -> float:
        """
        Compute the solvent-accessible surface area (SASA) score for a given protein structure.

        Parameters
        ----------
        pdb_file : Path
            The path to the PDB file of the protein structure.
        mutations : List[str], optional
            A list of mutations to be simulated on the protein structure. Only single mutations are supported.
            Pass no mutations if you want to compute the SASA of the wildtype.

        Returns
        -------
        float
            The computed SASA score.

        Raises
        ------
        AssertionError
            If the number of mutations is not 0 or 1.

        """
        if mutations is not None:
            assert len(mutations) in [
                0,
                1,
            ], "We only support single mutations for now. Pass no mutations if you want to compute the SASA of the wildtype."
        self._simulate_mutations(pdb_file, mutations)

        sasa_score = self._compute_sasa(pdb_file)
        return sasa_score

    def compute_stability_and_sasa(self, pdb_file: Path, mutations: List[str] = None):
        """Computes stability and sasa with a single foldx run,
        instead of two separate runs.

        Parameters
        ----------
        pdb_file : Path
            The path to the PDB file of the protein structure.
        mutations : List[str], optional
            A list of mutations to be simulated on the protein structure. Only single mutations are supported.
            Pass no mutations if you want to compute the SASA of the wildtype.
        """
        if mutations is not None:
            assert len(mutations) in [
                0,
                1,
            ], "We only support single mutations for now. Pass no mutations if you want to compute the energy and SASA of the wildtype."
        self._simulate_mutations(pdb_file, mutations)

        stability = -self._read_energy(pdb_file)
        sasa_score = self._compute_sasa(pdb_file)

        return stability, sasa_score

    def copy_foldx_files(self, pdb_file: Path):
        """Copies the pdb file to the working directory.

        Parameters
        ----------
        pdb_file : Path
            The path to the PDB file of the protein structure.
        """
        if (PATH_TO_FOLDX_FILES / "rotabase.txt").exists():
            # If rotabase exists, it's likely that the user is
            # using foldx v4. We should copy it if it's not
            # already in the working directory.
            try:
                os.symlink(
                    str(PATH_TO_FOLDX_FILES / "rotabase.txt"),
                    str(self.working_dir / "rotabase.txt"),
                )
            except FileExistsError:
                pass

        destination_path_for_pdb = self.working_dir / f"{pdb_file.stem}.pdb"
        if not destination_path_for_pdb.exists():
            shutil.copy(pdb_file, destination_path_for_pdb)

    @staticmethod
    def write_mutations_to_file(
        wildtype_resiudes: List[Residue], mutations: List[str], output_dir: Path
    ) -> None:
        """Writes the list of mutations to a file
        in the given directory.

        Parameters
        ----------
        wildtype_resiudes : List[Residue]
            The list of wildtype residues.
        mutations : List[str]
            The list of mutations to simulate.
        output_dir : Path
            The directory to write the file to.
        """
        # Write the mutations in the format of individual_list.txt
        lines = []
        for mutation in mutations:
            # e.g. mutations_in_line = ["MA0A", "AA5G"]
            # would represent two mutations of the same
            # wildtype: one at residue 0, which is mutated
            # from M to A; and one at residue 5, which is
            # mutated from A to G.
            # The A in between is the chain ID.
            mutations_in_line = mutations_from_wildtype_residues_and_mutant(
                wildtype_resiudes, mutation
            )

            # These are represented as a single line, separated
            # by commas, and terminated by a semicolon.
            lines.append(",".join(mutations_in_line) + ";")

        with open(output_dir / "individual_list.txt", "w") as f:
            f.writelines(lines)
